import io
import os

import psycopg
import pytest

import pgai
import pgai.semantic_catalog as semantic_catalog
import pgai.semantic_catalog.describe as describe
from pgai.semantic_catalog import file

from .utils import (
    PostgresContainer,
    load_airports,
)

DATABASE = "sc_02"

EMPTY_SET: set[str] = set()

ALL_TABLES: set[str] = {
    "account",
    "aircraft",
    "airport",
    "boarding_pass",
    "booking",
    "booking_leg",
    "events",
    "flight",
    "frequent_flyer",
    "hypertable_test",
    "passenger",
    "phone",
}

ALL_VIEWS: set[str] = {"events_daily", "flight_summary", "passenger_details"}

ALL_PROCS: set[str] = {"advance_air_time", "unsafe_sum", "update_flight_status"}


async def get_table_names(con: psycopg.AsyncConnection, oids: list[int]) -> set[str]:
    async with con.cursor() as cur:
        await cur.execute(
            """
            select distinct k.relname
            from pg_class k
            where k.oid = any(%s)
            and k.relkind in ('r', 'p', 'f')
        """,
            (oids,),
        )
        return {str(row[0]) for row in await cur.fetchall()}


async def get_view_names(con: psycopg.AsyncConnection, oids: list[int]) -> set[str]:
    async with con.cursor() as cur:
        await cur.execute(
            """
            select distinct k.relname
            from pg_class k
            where k.oid = any(%s)
            and k.relkind in ('v', 'm')
        """,
            (oids,),
        )
        return {str(row[0]) for row in await cur.fetchall()}


async def get_proc_names(con: psycopg.AsyncConnection, oids: list[int]) -> set[str]:
    async with con.cursor() as cur:
        await cur.execute(
            """
            select distinct p.proname
            from pg_proc p
            where p.oid = any(%s)
        """,
            (oids,),
        )
        return {str(row[0]) for row in await cur.fetchall()}


async def test_find_tables(container: PostgresContainer):
    tests: list[tuple[dict[str, str], set[str]]] = [
        ({"include_schema": "public"}, EMPTY_SET),
        ({"include_schema": "^public$"}, EMPTY_SET),
        ({"include_schema": "public$"}, EMPTY_SET),
        ({"include_schema": "^pub"}, EMPTY_SET),
        ({"include_schema": "postgres_air"}, ALL_TABLES),
        ({"include_schema": "^postgres_air$"}, ALL_TABLES),
        ({"include_schema": "_air$"}, ALL_TABLES),
        ({"exclude_schema": "public"}, ALL_TABLES),
        ({"exclude_schema": "^public$"}, ALL_TABLES),
        ({"exclude_schema": "public$"}, ALL_TABLES),
        ({"exclude_schema": "^pub"}, ALL_TABLES),
        ({"exclude_schema": "postgres_air"}, EMPTY_SET),
        ({"exclude_schema": "^postgres_air$"}, EMPTY_SET),
        ({"exclude_schema": "_air$"}, EMPTY_SET),
        ({"exclude_schema": "^pub", "include_schema": "_air$"}, ALL_TABLES),
        ({"exclude_schema": "_air$", "include_schema": "^pub"}, EMPTY_SET),
        ({"include_table": ".*"}, ALL_TABLES),
        ({"include_table": "(airport|aircraft)"}, {"airport", "aircraft"}),
        ({"exclude_table": "(airport|aircraft)"}, ALL_TABLES - {"airport", "aircraft"}),
        ({"include_table": "bob"}, EMPTY_SET),
        ({"include_table": "^booking"}, {"booking", "booking_leg"}),
        ({"exclude_schema": "_air$", "include_table": ".*"}, EMPTY_SET),
        ({"include_schema": "^postgres_air$", "exclude_table": ".*"}, EMPTY_SET),
        ({"exclude_schema": "^public$", "exclude_table": ".*"}, EMPTY_SET),
        (
            {"exclude_schema": "^public$", "include_table": "^a"},
            {"account", "aircraft", "airport"},
        ),
    ]
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database="postgres_air")
    ) as con:
        for test in tests:
            args, expected = test
            oids = await describe.find_tables(con, include_extensions=None, **args)
            actual = await get_table_names(con, oids)
            assert actual == expected, f"find_tables with {args} failed"


async def test_find_views(container: PostgresContainer):
    tests: list[tuple[dict[str, str], set[str]]] = [
        ({"include_schema": "public"}, EMPTY_SET),
        ({"include_schema": "^public$"}, EMPTY_SET),
        ({"include_schema": "public$"}, EMPTY_SET),
        ({"include_schema": "^pub"}, EMPTY_SET),
        ({"include_schema": "postgres_air"}, ALL_VIEWS),
        ({"include_schema": "^postgres_air$"}, ALL_VIEWS),
        ({"include_schema": "_air$"}, ALL_VIEWS),
        ({"exclude_schema": "public"}, ALL_VIEWS),
        ({"exclude_schema": "^public$"}, ALL_VIEWS),
        ({"exclude_schema": "public$"}, ALL_VIEWS),
        ({"exclude_schema": "^pub"}, ALL_VIEWS),
        ({"exclude_schema": "postgres_air"}, EMPTY_SET),
        ({"exclude_schema": "^postgres_air$"}, EMPTY_SET),
        ({"exclude_schema": "_air$"}, EMPTY_SET),
        ({"exclude_schema": "^pub", "include_schema": "_air$"}, ALL_VIEWS),
        ({"exclude_schema": "_air$", "include_schema": "^pub"}, EMPTY_SET),
        ({"include_view": ".*"}, ALL_VIEWS),
        (
            {"include_view": "(passenger_details|flight_summary|events_daily)"},
            ALL_VIEWS,
        ),
        (
            {"exclude_view": "(flight_summary|events_daily)"},
            {"passenger_details"},
        ),
        ({"include_view": "bob"}, EMPTY_SET),
        ({"include_view": "^flight"}, {"flight_summary"}),
        ({"exclude_schema": "_air$", "include_view": ".*"}, EMPTY_SET),
        ({"include_schema": "^postgres_air$", "exclude_view": ".*"}, EMPTY_SET),
        ({"exclude_schema": "^public$", "exclude_view": ".*"}, EMPTY_SET),
        ({"exclude_schema": "^public$", "include_view": "^f"}, {"flight_summary"}),
    ]
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database="postgres_air")
    ) as con:
        for test in tests:
            args, expected = test
            oids = await describe.find_views(con, include_extensions=None, **args)
            actual = await get_view_names(con, oids)
            assert actual == expected, f"find_views with {args} failed"


async def test_find_procs(container: PostgresContainer):
    tests: list[tuple[dict[str, str], set[str]]] = [
        ({"include_schema": "empty"}, EMPTY_SET),
        ({"include_schema": "^empty$"}, EMPTY_SET),
        ({"include_schema": "empty$"}, EMPTY_SET),
        ({"include_schema": "^emp"}, EMPTY_SET),
        ({"include_schema": "postgres_air"}, ALL_PROCS),
        ({"include_schema": "^postgres_air$"}, ALL_PROCS),
        ({"include_schema": "_air$"}, ALL_PROCS),
        ({"exclude_schema": "public"}, ALL_PROCS),
        ({"exclude_schema": "^public$"}, ALL_PROCS),
        ({"exclude_schema": "public$"}, ALL_PROCS),
        ({"exclude_schema": "^pub"}, ALL_PROCS),
        ({"exclude_schema": "(postgres_air|public)"}, EMPTY_SET),
        ({"exclude_schema": "^(postgres_air|public)$"}, EMPTY_SET),
        ({"exclude_schema": "^p"}, EMPTY_SET),
        ({"exclude_schema": "^emp", "include_schema": "_air$"}, ALL_PROCS),
        ({"exclude_schema": "_air$", "include_schema": "^emp"}, EMPTY_SET),
        ({"include_proc": "advance.*"}, {"advance_air_time"}),
        (
            {"include_proc": "(advance_air_time|unsafe_sum|update_flight_status)"},
            ALL_PROCS,
        ),
        (
            {"exclude_proc": "advance_air_time", "include_schema": "postgres_air"},
            {"update_flight_status", "unsafe_sum"},
        ),
        ({"include_proc": "bob"}, EMPTY_SET),
        ({"include_proc": "^advance"}, {"advance_air_time"}),
        ({"exclude_schema": "(postgres_air|public)", "include_proc": ".*"}, EMPTY_SET),
        ({"include_schema": "^postgres_air$", "exclude_proc": ".*"}, EMPTY_SET),
        ({"exclude_schema": "^public$", "exclude_proc": ".*"}, EMPTY_SET),
        (
            {"exclude_schema": "^public$", "include_proc": "^up"},
            {"update_flight_status"},
        ),
    ]
    async with (
        await psycopg.AsyncConnection.connect(
            container.connection_string(database="postgres_air")
        ) as con,
        con.cursor() as cur,
        con.transaction(force_rollback=True) as _,
    ):
        await cur.execute("create schema if not exists empty;")
        for test in tests:
            args, expected = test
            oids = await describe.find_procedures(con, include_extensions=None, **args)
            actual = await get_proc_names(con, oids)
            assert actual == expected, f"find_procs with {args} failed"


async def test_describe(container: PostgresContainer):
    if "ANTHROPIC_API_KEY" not in os.environ:
        pytest.skip("No anthroptic api key provided")
    expected = {
        "postgres_air.airport",
        "postgres_air.airport.airport_code",
        "postgres_air.airport.airport_name",
        "postgres_air.airport.city",
        "postgres_air.airport.airport_tz",
        "postgres_air.airport.continent",
        "postgres_air.airport.iso_country",
        "postgres_air.airport.iso_region",
        "postgres_air.airport.intnl",
        "postgres_air.airport.update_ts",
        "postgres_air.flight_summary",
        "postgres_air.flight_summary.flight_no",
        "postgres_air.flight_summary.departure_airport",
        "postgres_air.flight_summary.arrival_airport",
        "postgres_air.flight_summary.scheduled_departure",
        "postgres_air.advance_air_time",
    }
    buf = io.StringIO()
    async with (
        await psycopg.AsyncConnection.connect(
            container.connection_string(database="postgres_air")
        ) as tcon,
        tcon.transaction(force_rollback=True) as _,
    ):
        await load_airports(tcon)
        _ = await describe.describe(
            container.connection_string(database="postgres_air"),
            model="anthropic:claude-3-7-sonnet-latest",
            output=buf,
            include_schema="postgres_air",
            include_table="airport",
            include_view="flight_summary",
            include_proc="advance_air_time",
        )
    buf.seek(0)
    items = [item for item in file.import_from_yaml(buf)]
    assert len(items) == 3
    container.drop_database(DATABASE, force=True)
    container.create_database(DATABASE)
    pgai.install(container.connection_string(database=DATABASE))
    async with (
        await psycopg.AsyncConnection.connect(
            container.connection_string(database="postgres_air")
        ) as tcon,
        await psycopg.AsyncConnection.connect(
            container.connection_string(database=DATABASE)
        ) as ccon,
        ccon.cursor() as cur,
    ):
        sc = await semantic_catalog.create(ccon)
        await file.save_to_catalog(ccon, tcon, sc.id, iter(items))
        await cur.execute("""\
            select array_to_string(objnames, '.')
            from ai.semantic_catalog_obj_1
            where description is not null
        """)
        actual = {row[0] for row in await cur.fetchall()}
    assert actual == expected



import os

import psycopg
import pytest
from psycopg.rows import namedtuple_row

from db.tests.conftest import detailed_notice_handler


def db_url(user: str) -> str:
    return f"postgres://{user}@127.0.0.1:5432/test"


@pytest.mark.skipif(
    os.getenv("PG_MAJOR") == "15", reason="extension does not support pg15"
)
def test_same_table_vectorizer_timescaledb():
    with psycopg.connect(
        db_url("postgres"), autocommit=True, row_factory=namedtuple_row
    ) as con:
        with con.cursor() as cur:
            cur.execute("create extension if not exists timescaledb")
            cur.execute("select to_regrole('bob') is null")
            if cur.fetchone()[0] is True:
                cur.execute("create user bob")
            cur.execute("select to_regrole('adelaide') is null")
            if cur.fetchone()[0] is True:
                cur.execute("create user adelaide")
            cur.execute("create extension if not exists ai cascade")
    with psycopg.connect(
        db_url("test"), autocommit=True, row_factory=namedtuple_row
    ) as con:
        con.add_notice_handler(detailed_notice_handler)
        with con.cursor() as cur:
            cur.execute("drop schema if exists website cascade")
            cur.execute("create schema website")
            cur.execute("drop table if exists website.blog")
            cur.execute("""
                create table website.blog
                ( id int not null generated always as identity
                , title text not null
                , published timestamptz
                , body text not null
                , drop_me text
                , primary key (title, published)
                )
            """)
            cur.execute(
                """grant select, insert, update, delete on website.blog to bob, adelaide"""
            )
            cur.execute("""grant usage on schema website to adelaide""")
            cur.execute("""
                insert into website.blog(title, published, body)
                values
                  ('how to cook a hot dog', '2024-01-06'::timestamptz, 'put it on a hot grill')
                , ('how to make a sandwich', '2023-01-06'::timestamptz, 'put a slice of meat between two pieces of bread')
                , ('how to make stir fry', '2022-01-06'::timestamptz, 'pick up the phone and order takeout')
            """)

            # drop the drop_me column
            cur.execute("alter table website.blog drop column drop_me")

            # create a vectorizer for the blog table
            # language=PostgreSQL
            cur.execute("""
            select ai.create_vectorizer
            ( 'website.blog'::regclass
            , loading=>ai.loading_column('body')
            , embedding=>ai.embedding_openai('text-embedding-3-small', 768)
            , formatting=>ai.formatting_python_template('title: $title published: $published $chunk')
            , scheduling=>ai.scheduling_timescaledb
                    ( interval '5m'
                    , initial_start=>'2050-01-06'::timestamptz
                    , timezone=>'America/Chicago'
                    )
            , destination=>ai.destination_column('embedding1')
            , chunking=>ai.chunking_none()
            );
            """)
            vectorizer_id = cur.fetchone()[0]

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer_id,))
            actual = cur.fetchone()[0]
            assert actual == 3

            cur.execute(
                """
                select (x.config->'scheduling'->>'job_id')::int 
                from ai.vectorizer x 
                where x.id = %s
                """,
                (vectorizer_id,),
            )
            job_id = cur.fetchone()[0]

            cur.execute("call public.run_job(%s)", (job_id,))

            # check that the queue has 0 rows
            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer_id,))
            actual = cur.fetchone()[0]
            assert actual == 0

            # assert website table has embedding column
            cur.execute(
                "select column_name from information_schema.columns where table_name = 'blog'"
            )
            columns = [row[0] for row in cur.fetchall()]
            assert "embedding1" in columns

            # edit website rows
            cur.execute("""
                update website.blog
                set title = 'how to cook a cold dog'
                where title = 'how to cook a hot dog'
            """)

            # check that the queue has 0 rows
            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer_id,))
            actual = cur.fetchone()[0]
            assert actual == 1

            cur.execute("call public.run_job(%s)", (job_id,))
            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer_id,))
            actual = cur.fetchone()[0]
            assert actual == 0

            # edit the embedding column with a vector
            cur.execute("""
                update website.blog
                set embedding1 = array_fill(0.1::float8, ARRAY[768])::vector
                where title = 'how to cook a cold dog'
            """)

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer_id,))
            actual = cur.fetchone()[0]
            assert actual == 0

            # Create a second vectorizer

            cur.execute("""
                select ai.create_vectorizer
                ( 'website.blog'::regclass
                , loading=>ai.loading_column('body')
                , embedding=>ai.embedding_openai('text-embedding-3-small', 768)
                , formatting=>ai.formatting_python_template('title: $title published: $published $chunk')
                , scheduling=>ai.scheduling_timescaledb
                        ( interval '5m'
                        , initial_start=>'2050-01-06'::timestamptz
                        , timezone=>'America/Chicago'
                        )
                , destination=>ai.destination_column('embedding2')
                , chunking=>ai.chunking_none()
                );
                """)
            vectorizer2_id = cur.fetchone()[0]

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer2_id,))
            actual = cur.fetchone()[0]
            assert actual == 3

            cur.execute(
                """
                select (x.config->'scheduling'->>'job_id')::int 
                from ai.vectorizer x 
                where x.id = %s
                """,
                (vectorizer2_id,),
            )
            job2_id = cur.fetchone()[0]

            cur.execute("call public.run_job(%s)", (job2_id,))

            # check that the queue has 0 rows
            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer2_id,))
            actual = cur.fetchone()[0]
            assert actual == 0

            # assert website table has embedding2 column
            cur.execute(
                "select column_name from information_schema.columns where table_name = 'blog'"
            )
            columns = [row[0] for row in cur.fetchall()]
            assert "embedding2" in columns

            # edit website rows
            cur.execute("""
                update website.blog
                set title = 'how to cook a warm dog'
                where title = 'how to cook a cold dog'
            """)

            # check that both queues have 1 row
            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer2_id,))
            actual = cur.fetchone()[0]
            assert actual == 1

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer_id,))
            actual = cur.fetchone()[0]
            assert actual == 1

            cur.execute("call public.run_job(%s)", (job2_id,))
            cur.execute("call public.run_job(%s)", (job_id,))

            # edit the embedding columns
            cur.execute("""
                update website.blog
                set embedding1 = array_fill(0.2::float8, ARRAY[768])::vector
                where title = 'how to cook a warm dog'
            """)

            cur.execute("""
                update website.blog
                set embedding2 = array_fill(0.3::float8, ARRAY[768])::vector
                where title = 'how to cook a warm dog'
            """)

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer2_id,))
            actual = cur.fetchone()[0]
            assert actual == 0

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer_id,))
            actual = cur.fetchone()[0]
            assert actual == 0

            # Create a third non-same-table vectorizer

            cur.execute("""
                           select ai.create_vectorizer
                           ( 'website.blog'::regclass
                           , loading=>ai.loading_column('body')
                           , embedding=>ai.embedding_openai('text-embedding-3-small', 768)
                           , formatting=>ai.formatting_python_template('title: $title published: $published $chunk')
                           , scheduling=>ai.scheduling_timescaledb
                                   ( interval '5m'
                                   , initial_start=>'2050-01-06'::timestamptz
                                   , timezone=>'America/Chicago'
                                   )
                           );
                           """)

            vectorizer3_id = cur.fetchone()[0]

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer3_id,))
            actual = cur.fetchone()[0]
            assert actual == 3

            cur.execute(
                """
                select (x.config->'scheduling'->>'job_id')::int 
                from ai.vectorizer x 
                where x.id = %s
                """,
                (vectorizer3_id,),
            )
            job3_id = cur.fetchone()[0]

            cur.execute("call public.run_job(%s)", (job3_id,))

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer3_id,))
            actual = cur.fetchone()[0]
            assert actual == 0

            cur.execute("""
                            update website.blog
                            set title = 'how to cook a spicy dog'
                            where title = 'how to cook a warm dog'
                        """)

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer_id,))
            actual = cur.fetchone()[0]
            assert actual == 1

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer2_id,))
            actual = cur.fetchone()[0]
            assert actual == 1

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer3_id,))
            actual = cur.fetchone()[0]
            assert actual == 1

            cur.execute("call public.run_job(%s)", (job_id,))
            cur.execute("call public.run_job(%s)", (job2_id,))
            cur.execute("call public.run_job(%s)", (job3_id,))

            cur.execute("""
                            update website.blog
                            set embedding1 = array_fill(0.2::float8, ARRAY[768])::vector
                            where title = 'how to make a sandwich'
                        """)

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer_id,))
            actual = cur.fetchone()[0]
            assert actual == 0

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer2_id,))
            actual = cur.fetchone()[0]
            assert actual == 0

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer3_id,))
            actual = cur.fetchone()[0]
            assert actual == 0

            # Do an insert
            cur.execute("""
                insert into website.blog(title, published, body)
                values
                  ('how to cook spaghetti', '2024-01-06'::timestamptz, 'put it on a hot grill')
            """)

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer_id,))
            actual = cur.fetchone()[0]
            assert actual == 1

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer2_id,))
            actual = cur.fetchone()[0]
            assert actual == 1

            cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer3_id,))
            actual = cur.fetchone()[0]
            assert actual == 1

            # Do a delete
            cur.execute("""
                delete from website.blog
                where title = 'how to cook spaghetti'
            """)



import importlib
import pkgutil
import sys
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

import semver
import structlog

logger = structlog.get_logger()

# config generic type
C = TypeVar("C")
MigrationFunc = Callable[[C], dict[str, Any]]


@dataclass
class Migration(Generic[C]):
    version: str  # target version
    source_vectorizer_class: type[C]  # config dataclass for the source version
    apply: MigrationFunc[C]
    description: str = ""


# registry to hold all migrations
migrations: list[Migration[Any]] = []

MigrationDecorator = Callable[[MigrationFunc[C]], MigrationFunc[C]]


def register_migration(
    version: str, source_vectorizer_class: type[C], description: str = ""
) -> MigrationDecorator[C]:
    """
    Decorator to register a migration function with
    config class validation
    """

    def decorator(func: MigrationFunc[C]) -> MigrationFunc[C]:
        migrations.append(
            Migration(
                version=version,
                source_vectorizer_class=source_vectorizer_class,
                apply=func,
                description=description,
            )
        )
        migrations.sort(key=lambda m: semver.VersionInfo.parse(m.version))
        return func

    return decorator


def get_latest_version() -> str:
    """Get the latest migration version available"""
    if not migrations:
        return "0.0.0"
    return migrations[-1].version


def apply_migrations(data: dict[str, Any]) -> dict[str, Any]:
    """
    Apply all necessary migrations incrementally until reaching the latest version

    Args:
        data: The config data to migrate

    Returns:
        The migrated config data
    """
    if not data:
        return data

    # determine starting version
    config_dict = data.get("config")
    if config_dict is None:
        logger.warning("Unable to migrate configuration: 'config' field missing")
        return data
    original_version = config_dict.get("version")
    if original_version is None:
        logger.warning("Unable to migrate configuration: 'version' field missing")
        return data

    original_version = original_version.replace("-dev", "")

    current = semver.VersionInfo.parse(original_version)
    latest = semver.VersionInfo.parse(get_latest_version())

    # no migrations needed if already at latest
    if current >= latest:
        return data

    result = dict(data)

    # only migrations with a version greater than the current version are applicable
    applicable_migrations = [
        m for m in migrations if current < semver.VersionInfo.parse(m.version)
    ]

    current_version = original_version
    # migrations are applied sequentially
    for migration in applicable_migrations:
        logger.info(
            f"Applying Vectorizer Config migration "
            f"from version {current_version} "
            f"to version {migration.version}: {migration.description}"
        )

        # instantiate config class for this version so we can validate the input
        vectorizer_instance = migration.source_vectorizer_class(**result)

        result = migration.apply(vectorizer_instance)

        # updating the version after each successful migration so we
        # don't re-apply migrations
        current_version = migration.version
        result["config"]["version"] = migration.version

    result["config"]["original_version"] = original_version

    return result


def load_migrations():
    """
    dynamically import all migration modules to register them
    through the register_migration decorator
    """
    package = sys.modules[__name__]
    modules = pkgutil.iter_modules(package.__path__, package.__name__ + ".")
    for _, name, is_pkg in modules:
        if not is_pkg:
            importlib.import_module(name)


load_migrations()



"""Database object reference fixing utilities for semantic catalog.

This module provides functions to fix database object references in the semantic
catalog. When database objects like tables, views, or columns are changed, their
internal identifiers and object names might need to be updated in the semantic catalog
to maintain proper references.
"""

import logging
from dataclasses import dataclass

import psycopg
from psycopg.rows import class_row
from psycopg.sql import SQL, Composed, Identifier, Literal
from rich.console import Console
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
)

logger = logging.getLogger(__name__)


@dataclass
class _Object:
    """Represents a database object reference stored in the semantic catalog.

    Contains both PostgreSQL internal identifiers (classid, objid, objsubid) and
    object name identifiers (objtype, objnames, objargs) for a database object.
    """

    id: int  # Semantic catalog object ID
    classid: int  # PostgreSQL pg_class.oid for the system catalog containing the object
    objid: int  # PostgreSQL oid of the object in the system catalog
    objsubid: int  # Sub-object ID for columns and other sub-objects
    objtype: str  # Type of object (table, view, function, etc.)
    objnames: list[str]  # Fully qualified object name as a list of identifiers
    objargs: list[str]  # Function arguments (if applicable)


@dataclass
class _Ids:
    """Container for PostgreSQL internal object identifiers.

    Represents the system identifiers for a PostgreSQL database object,
    as returned by pg_get_object_address().
    """

    classid: int  # PostgreSQL pg_class.oid for the system catalog containing the object
    objid: int  # PostgreSQL oid of the object in the system catalog
    objsubid: int  # Sub-object ID for columns and other sub-objects


@dataclass
class _Names:
    """Container for PostgreSQL object name identifiers.

    Represents the name-based identifiers for a PostgreSQL database object,
    as returned by pg_identify_object_as_address().
    """

    objtype: str  # Type of object (table, view, function, etc.)
    objnames: list[str]  # Fully qualified object name as a list of identifiers
    objargs: list[str]  # Function arguments (if applicable)


async def _list_ids(catalog_con: psycopg.AsyncConnection, catalog_id: int) -> list[int]:
    """Retrieve all object IDs from the specified semantic catalog.

    Args:
        catalog_con: Connection to the database containing the semantic catalog
        catalog_id: ID of the semantic catalog

    Returns:
        List of all object IDs in the catalog, sorted by ID
    """
    async with catalog_con.cursor() as cur:
        await cur.execute(
            SQL("""\
            select id
            from ai.{}
            order by id
        """).format(Identifier(f"semantic_catalog_obj_{catalog_id}"))
        )
        return [row[0] for row in await cur.fetchall()]


async def _get_obj(
    catalog_con: psycopg.AsyncConnection, catalog_id: int, id: int
) -> _Object | None:
    """Retrieve a specific object from the semantic catalog by ID.

    Args:
        catalog_con: Connection to the database containing the semantic catalog
        catalog_id: ID of the semantic catalog
        id: ID of the specific object to retrieve

    Returns:
        The object if found, None otherwise
    """
    async with catalog_con.cursor(row_factory=class_row(_Object)) as cur:
        await cur.execute(
            SQL("""\
            select
              id
            , classid
            , objid
            , objsubid
            , objtype
            , objnames
            , objargs
            from ai.{}
            where id = %s
        """).format(Identifier(f"semantic_catalog_obj_{catalog_id}")),
            (id,),
        )
        obj = await cur.fetchone()
        return obj


async def _ids_from_names(
    target_con: psycopg.AsyncConnection, obj: _Object
) -> _Ids | None:
    """Convert object name identifiers to PostgreSQL internal IDs.

    Uses pg_get_object_address() to find current internal IDs for an object
    identified by name, type, and arguments.

    Args:
        target_con: Connection to the target database
        obj: Object with name identifiers to convert

    Returns:
        Internal IDs if object exists, None if object not found
    """
    try:
        async with (
            target_con.cursor(row_factory=class_row(_Ids)) as cur,
            target_con.transaction() as _,
        ):
            await cur.execute(
                """\
                select
                  a.classid
                , a.objid
                , a.objsubid as objsubid
                from pg_get_object_address
                ( case
                    when %(objtype)s = 'view column' then 'table column'
                    else %(objtype)s
                  end
                , %(objnames)s
                , %(objargs)s
                ) a
            """,
                dict(objtype=obj.objtype, objnames=obj.objnames, objargs=obj.objargs),
            )
            ids = await cur.fetchone()
            return ids
    except psycopg.Error as e:
        logger.debug(f"object not found {obj.id}: {str(e)}")
    return None


def _del_obj_sql(catalog_id: int, obj: _Object) -> Composed:
    """Generate SQL to delete an object from the semantic catalog.

    Args:
        catalog_id: ID of the semantic catalog
        obj: Object to delete

    Returns:
        SQL query to delete the object
    """
    return SQL("DELETE from ai.{table} WHERE id = {id};").format(
        table=Identifier(f"semantic_catalog_obj_{catalog_id}"),
        id=Literal(obj.id),
    )


def _ids_dont_match(obj: _Object, ids: _Ids) -> bool:
    """Check if the internal IDs in the object match the provided IDs.

    Args:
        obj: Object with stored IDs
        ids: Current IDs to compare against

    Returns:
        True if IDs don't match, False if they do
    """
    return (
        obj.classid != ids.classid
        or obj.objid != ids.objid
        or obj.objsubid != ids.objsubid
    )


def _update_ids_sql(catalog_id: int, obj: _Object, ids: _Ids) -> Composed:
    """Generate SQL to update an object's internal IDs in the semantic catalog.

    Args:
        catalog_id: ID of the semantic catalog
        obj: Object to update
        ids: New internal IDs to set

    Returns:
        SQL query to update the object's IDs
    """
    return SQL(
        "UPDATE ai.{table} SET classid={classid}, objid={objid}, objsubid={objsubid} WHERE id = {id};"
    ).format(
        table=Identifier(f"semantic_catalog_obj_{catalog_id}"),
        classid=Literal(ids.classid),
        objid=Literal(ids.objid),
        objsubid=Literal(ids.objsubid),
        id=Literal(obj.id),
    )


async def _names_from_ids(
    target_con: psycopg.AsyncConnection, obj: _Object
) -> _Names | None:
    """Convert PostgreSQL internal IDs to object name identifiers.

    Uses pg_identify_object_as_address() to find current name identifiers for an object
    identified by internal IDs.

    Args:
        target_con: Connection to the target database
        obj: Object with internal IDs to convert

    Returns:
        Name identifiers if object exists, None if object not found
    """
    try:
        async with (
            target_con.cursor(row_factory=class_row(_Names)) as cur,
            target_con.transaction() as _,
        ):
            await cur.execute(
                """\
                select
                  a."type" as objtype
                , a.object_names as objnames
                , a.object_args as objargs
                from pg_identify_object_as_address
                ( %(classid)s
                , %(objid)s
                , %(objsubid)s
                ) a
            """,
                dict(classid=obj.classid, objid=obj.objid, objsubid=obj.objsubid),
            )
            names = await cur.fetchone()
            return names
    except psycopg.Error as e:
        logger.debug(f"object not found {obj.id}: {str(e)}")
    return None


def _names_dont_match(obj: _Object, names: _Names) -> bool:
    """Check if the name identifiers in the object match the provided names.

    Args:
        obj: Object with stored name identifiers
        names: Current name identifiers to compare against

    Returns:
        True if name identifiers don't match, False if they do
    """
    return (
        obj.objtype != names.objtype
        or obj.objnames != names.objnames
        or obj.objargs != names.objargs
    )


def _update_names_sql(catalog_id: int, obj: _Object, names: _Names) -> Composed:
    """Generate SQL to update an object's name identifiers in the semantic catalog.

    Args:
        catalog_id: ID of the semantic catalog
        obj: Object to update
        names: New name identifiers to set

    Returns:
        SQL query to update the object's name identifiers
    """
    return SQL(
        "UPDATE ai.{table} set objtype={objtype}, objnames={objnames}, objargs={objargs} WHERE id = {id};"
    ).format(
        table=Identifier(f"semantic_catalog_obj_{catalog_id}"),
        objtype=Literal(names.objtype),
        objnames=Literal(names.objnames),
        objargs=Literal(names.objargs),
        id=Literal(obj.id),
    )


async def _defer_name_constraint(
    catalog_cur: psycopg.AsyncCursor, catalog_id: int
) -> None:
    await catalog_cur.execute(
        """\
        select x.conname
        from pg_class k
        inner join pg_namespace n on (k.relnamespace = n.oid)
        cross join lateral
        (
            select array_agg(a.attnum order by a.attnum) as cols
            from pg_attribute a
            where a.attrelid = k.oid
            and a.attname in ('objtype', 'objnames', 'objargs')
        ) a
        inner join pg_constraint x on (x.conrelid = k.oid and x.conkey = a.cols)
        where k.relname = %s
        and n.nspname = 'ai'
        and x.contype = 'u'
        and x.condeferrable
    """,
        (f"semantic_catalog_obj_{catalog_id}",),
    )
    row = await catalog_cur.fetchone()
    assert row is not None, "could not find unique constraint"
    conname: str = str(row[0])
    await catalog_cur.execute(
        SQL("""\
        set constraints ai.{conname} deferred
    """).format(conname=Identifier(conname))
    )


async def _defer_id_constraint(
    catalog_cur: psycopg.AsyncCursor, catalog_id: int
) -> None:
    await catalog_cur.execute(
        """\
        select x.conname
        from pg_class k
        inner join pg_namespace n on (k.relnamespace = n.oid)
        cross join lateral
        (
            select array_agg(a.attnum order by a.attnum) as cols
            from pg_attribute a
            where a.attrelid = k.oid
            and a.attname in ('classid', 'objid', 'objsubid')
        ) a
        inner join pg_constraint x on (x.conrelid = k.oid and x.conkey = a.cols)
        where k.relname = %s
        and n.nspname = 'ai'
        and x.contype = 'u'
        and x.condeferrable
    """,
        (f"semantic_catalog_obj_{catalog_id}",),
    )
    row = await catalog_cur.fetchone()
    assert row is not None, "could not find unique constraint"
    conname: str = str(row[0])
    await catalog_cur.execute(
        SQL("""\
        set constraints ai.{conname} deferred
    """).format(conname=Identifier(conname))
    )


async def fix_ids(
    catalog_con: psycopg.AsyncConnection,
    target_con: psycopg.AsyncConnection,
    catalog_id: int,
    dry_run: bool,
    console: Console,
) -> None:
    """Fix internal PostgreSQL IDs in the semantic catalog.

    Checks all objects in the semantic catalog against the target database. For each object:
    - If object doesn't exist in target database, marks it for deletion
    - If object's internal IDs don't match current values, marks it for update
    - If object is already correct, leaves it unchanged

    If not a dry run, performs all deletions and updates in a transaction.
    Shows progress with a rich progress bar.

    Args:
        catalog_con: Connection to the database containing the semantic catalog
        target_con: Connection to the target database containing the actual objects
        catalog_id: ID of the semantic catalog to fix
        dry_run: If True, only check for issues without making changes
        console: Rich console for output and progress display
    """
    pbcols = [
        SpinnerColumn(),
        TextColumn("{task.description}"),
        BarColumn(),
        MofNCompleteColumn(),
        TimeElapsedColumn(),
    ]

    with Progress(*pbcols, console=console) as progress:
        ids = await _list_ids(catalog_con, catalog_id)

        task_check = progress.add_task(
            "checking ids",
            total=len(ids),
            visible=len(ids) > 0,
        )

        queries: list[tuple[str, Composed]] = []
        for id in ids:
            obj = await _get_obj(catalog_con, catalog_id, id)
            if obj is None:
                progress.update(task_check, advance=1.0)
                continue
            ids = await _ids_from_names(target_con, obj)
            if ids is None:
                logger.info(f"obj {obj.id} no longer matches anything. deleting")
                progress.console.print(f'{obj.id} {".".join(obj.objnames)}: MISSING')
                queries.append(
                    (
                        "DELETING: " + ".".join(obj.objnames),
                        _del_obj_sql(catalog_id, obj),
                    )
                )
            elif _ids_dont_match(obj, ids):
                logger.info(f"updating the ids for obj {obj.id}")
                progress.console.print(f'{obj.id} {".".join(obj.objnames)}: FIX')
                queries.append(
                    (
                        "UPDATING: " + ".".join(obj.objnames),
                        _update_ids_sql(catalog_id, obj, ids),
                    )
                )
            else:
                progress.console.print(f'{obj.id} {".".join(obj.objnames)}: OK')
                logger.info(f"obj {obj.id} is already correct")
            progress.update(task_check, advance=1.0)

        task_fix = progress.add_task(
            "applying fixes",
            total=len(queries),
            visible=len(queries) > 0,
        )

        if dry_run:
            return

        # do these changes in a transaction to take advantage of deferrable constraints
        async with catalog_con.transaction() as _, catalog_con.cursor() as cur:
            await _defer_id_constraint(cur, catalog_id)
            for msg, query in queries:
                progress.console.print(msg)
                await cur.execute(query)
                progress.update(task_fix, advance=1.0)


async def fix_names(
    catalog_con: psycopg.AsyncConnection,
    target_con: psycopg.AsyncConnection,
    catalog_id: int,
    dry_run: bool,
    console: Console,
) -> None:
    """Fix object name identifiers in the semantic catalog.

    Checks all objects in the semantic catalog against the target database. For each object:
    - If object doesn't exist in target database, marks it for deletion
    - If object's name identifiers don't match current values, marks it for update
    - If object is already correct, leaves it unchanged

    If not a dry run, performs all deletions and updates in a transaction.
    Shows progress with a rich progress bar.

    Args:
        catalog_con: Connection to the database containing the semantic catalog
        target_con: Connection to the target database containing the actual objects
        catalog_id: ID of the semantic catalog to fix
        dry_run: If True, only check for issues without making changes
        console: Rich console for output and progress display
    """
    pbcols = [
        SpinnerColumn(),
        TextColumn("{task.description}"),
        BarColumn(),
        MofNCompleteColumn(),
        TimeElapsedColumn(),
    ]

    with Progress(*pbcols, console=console) as progress:
        ids = await _list_ids(catalog_con, catalog_id)

        task_check = progress.add_task(
            "checking names",
            total=len(ids),
            visible=len(ids) > 0,
        )

        queries: list[tuple[str, Composed]] = []
        for id in ids:
            obj = await _get_obj(catalog_con, catalog_id, id)
            if obj is None:
                progress.update(task_check, advance=1.0)
                continue
            names = await _names_from_ids(target_con, obj)
            if names is None:
                logger.info(f"obj {obj.id} no longer matches anything. deleting")
                progress.console.print(f'{obj.id} {".".join(obj.objnames)}: MISSING')
                queries.append(
                    (
                        "DELETING: " + ".".join(obj.objnames),
                        _del_obj_sql(catalog_id, obj),
                    )
                )
            elif _names_dont_match(obj, names):
                logger.info(f"updating the names for obj {obj.id}")
                progress.console.print(f'{obj.id} {".".join(obj.objnames)}: FIX')
                queries.append(
                    (
                        "UPDATING: " + ".".join(obj.objnames),
                        _update_names_sql(catalog_id, obj, names),
                    )
                )
            else:
                progress.console.print(f'{obj.id} {".".join(obj.objnames)}: OK')
                logger.info(f"obj {obj.id} is already correct")
            progress.update(task_check, advance=1.0)

        task_fix = progress.add_task(
            "applying fixes",
            total=len(queries),
            visible=len(queries) > 0,
        )

        if dry_run:
            return

        # do these changes in a transaction to take advantage of deferrable constraints
        async with catalog_con.transaction() as _, catalog_con.cursor() as cur:
            await _defer_name_constraint(cur, catalog_id)
            for msg, query in queries:
                progress.console.print(msg)
                await cur.execute(query)
                progress.update(task_fix, advance=1.0)



import os

import psycopg
import pytest

# skip tests in this module if disabled
enable_vectorizer_tests = os.getenv("ENABLE_VECTORIZER_TESTS")
if enable_vectorizer_tests == "0":
    pytest.skip(allow_module_level=True)


def db_url(user: str) -> str:
    return f"postgres://{user}@127.0.0.1:5432/test"


def test_indexing_none():
    tests = [
        (
            "select ai.indexing_none()",
            {
                "implementation": "none",
                "config_type": "indexing",
            },
        ),
    ]
    with psycopg.connect(db_url("test")) as con:
        with con.cursor() as cur:
            for query, expected in tests:
                cur.execute(query)
                actual = cur.fetchone()[0]
                assert actual.keys() == expected.keys()
                for k, v in actual.items():
                    assert k in expected and v == expected[k]


def test_indexing_diskann():
    tests = [
        (
            "select ai.indexing_diskann()",
            {
                "implementation": "diskann",
                "config_type": "indexing",
                "min_rows": 100_000,
                "create_when_queue_empty": True,
            },
        ),
        (
            "select ai.indexing_diskann(min_rows=>500)",
            {
                "implementation": "diskann",
                "config_type": "indexing",
                "min_rows": 500,
                "create_when_queue_empty": True,
            },
        ),
        (
            "select ai.indexing_diskann(storage_layout=>'plain')",
            {
                "implementation": "diskann",
                "config_type": "indexing",
                "min_rows": 100_000,
                "storage_layout": "plain",
                "create_when_queue_empty": True,
            },
        ),
        (
            "select ai.indexing_diskann(storage_layout=>'plain', create_when_queue_empty=>false)",
            {
                "implementation": "diskann",
                "config_type": "indexing",
                "min_rows": 100_000,
                "storage_layout": "plain",
                "create_when_queue_empty": False,
            },
        ),
        (
            """
            select ai.indexing_diskann
            ( storage_layout=>'memory_optimized'
            , num_neighbors=>50
            , search_list_size=>150
            , max_alpha=>1.2
            , num_dimensions=>768
            , num_bits_per_dimension=>2
            )
            """,
            {
                "implementation": "diskann",
                "config_type": "indexing",
                "min_rows": 100_000,
                "storage_layout": "memory_optimized",
                "num_neighbors": 50,
                "search_list_size": 150,
                "max_alpha": 1.2,
                "num_dimensions": 768,
                "num_bits_per_dimension": 2,
                "create_when_queue_empty": True,
            },
        ),
        (
            """
            select ai.indexing_diskann
            ( storage_layout=>'memory_optimized'
            , num_neighbors=>50
            , search_list_size=>150
            , max_alpha=>1.2
            , num_dimensions=>768
            , num_bits_per_dimension=>2
            , create_when_queue_empty=>false
            )
            """,
            {
                "implementation": "diskann",
                "config_type": "indexing",
                "min_rows": 100_000,
                "storage_layout": "memory_optimized",
                "num_neighbors": 50,
                "search_list_size": 150,
                "max_alpha": 1.2,
                "num_dimensions": 768,
                "num_bits_per_dimension": 2,
                "create_when_queue_empty": False,
            },
        ),
    ]
    with psycopg.connect(db_url("test")) as con:
        with con.cursor() as cur:
            for query, expected in tests:
                cur.execute(query)
                actual = cur.fetchone()[0]
                assert actual.keys() == expected.keys()
                for k, v in actual.items():
                    assert k in expected and v == expected[k]


def test_indexing_hnsw():
    tests = [
        (
            "select ai.indexing_hnsw()",
            {
                "implementation": "hnsw",
                "config_type": "indexing",
                "min_rows": 100_000,
                "opclass": "vector_cosine_ops",
                "create_when_queue_empty": True,
            },
        ),
        (
            "select ai.indexing_hnsw(min_rows=>500)",
            {
                "implementation": "hnsw",
                "config_type": "indexing",
                "min_rows": 500,
                "opclass": "vector_cosine_ops",
                "create_when_queue_empty": True,
            },
        ),
        (
            "select ai.indexing_hnsw(opclass=>'vector_l1_ops')",
            {
                "implementation": "hnsw",
                "config_type": "indexing",
                "min_rows": 100_000,
                "opclass": "vector_l1_ops",
                "create_when_queue_empty": True,
            },
        ),
        (
            "select ai.indexing_hnsw(m=>10, ef_construction=>100)",
            {
                "implementation": "hnsw",
                "config_type": "indexing",
                "min_rows": 100_000,
                "m": 10,
                "ef_construction": 100,
                "opclass": "vector_cosine_ops",
                "create_when_queue_empty": True,
            },
        ),
        (
            "select ai.indexing_hnsw(m=>10, ef_construction=>100, create_when_queue_empty=>false)",
            {
                "implementation": "hnsw",
                "config_type": "indexing",
                "min_rows": 100_000,
                "m": 10,
                "ef_construction": 100,
                "opclass": "vector_cosine_ops",
                "create_when_queue_empty": False,
            },
        ),
    ]
    with psycopg.connect(db_url("test")) as con:
        with con.cursor() as cur:
            for query, expected in tests:
                cur.execute(query)
                actual = cur.fetchone()[0]
                assert actual.keys() == expected.keys()
                for k, v in actual.items():
                    assert k in expected and v == expected[k]


def test_validate_indexing():
    ok = [
        "select ai._validate_indexing(ai.indexing_none())",
        "select ai._validate_indexing(ai.indexing_hnsw())",
        "select ai._validate_indexing(ai.indexing_hnsw(opclass=>'vector_ip_ops'))",
        "select ai._validate_indexing(ai.indexing_hnsw(opclass=>'vector_cosine_ops'))",
        "select ai._validate_indexing(ai.indexing_hnsw(opclass=>'vector_l1_ops'))",
        "select ai._validate_indexing(ai.indexing_hnsw(opclass=>null))",
        "select ai._validate_indexing(ai.indexing_hnsw(create_when_queue_empty=>false))",
        "select ai._validate_indexing(ai.indexing_diskann())",
        "select ai._validate_indexing(ai.indexing_diskann(storage_layout=>'plain'))",
        "select ai._validate_indexing(ai.indexing_diskann(storage_layout=>'memory_optimized'))",
        "select ai._validate_indexing(ai.indexing_diskann(storage_layout=>null))",
        "select ai._validate_indexing(ai.indexing_diskann(create_when_queue_empty=>false))",
    ]
    bad = [
        (
            "select ai._validate_indexing(ai.indexing_hnsw(opclass=>'peter'))",
            "invalid opclass",
        ),
        (
            "select ai._validate_indexing(ai.indexing_diskann(storage_layout=>'super_advanced'))",
            "invalid storage",
        ),
        (
            "select ai._validate_indexing(ai.scheduling_none())",
            "invalid config_type for indexing config",
        ),
    ]
    with psycopg.connect(db_url("test"), autocommit=True) as con:
        with con.cursor() as cur:
            for query in ok:
                cur.execute(query)
                assert True
            for query, err in bad:
                try:
                    cur.execute(query)
                except psycopg.ProgrammingError as ex:
                    msg = str(ex.args[0])
                    assert len(msg) >= len(err) and msg[: len(err)] == err
                else:
                    pytest.fail(f"expected exception: {err}")



"""Ollama embedding provider for the semantic catalog vectorizer.

This module implements embedding functionality using Ollama as the embedding provider.
It provides functions for embedding both batches of content and individual queries.
"""

from collections.abc import Sequence

from ollama import AsyncClient

from pgai.semantic_catalog.vectorizer import OllamaConfig
from pgai.semantic_catalog.vectorizer.models import EmbedRow


async def embed_batch(config: OllamaConfig, batch: list[EmbedRow]) -> None:
    """Generate embeddings for a batch of content using Ollama.

    Creates vector embeddings for multiple items using the Ollama API and
    updates the vector field in each EmbedRow object with the resulting embedding.

    Args:
        config: Configuration for the Ollama embedding service.
        batch: List of EmbedRow objects containing content to be embedded.

    Raises:
        RuntimeError: If the number of embeddings returned doesn't match the batch size.
    """
    client = AsyncClient(host=config.base_url)  # TODO: cache this?
    response = await client.embed(config.model, [x.content for x in batch])
    if len(response.embeddings) != len(batch):
        raise RuntimeError(
            f"{len(batch)} items sent to ollama but {len(response.embeddings)} embeddings returned"
        )
    for i, embed in enumerate(response.embeddings):
        batch[i].vector = embed


async def embed_query(config: OllamaConfig, query: str) -> Sequence[float]:
    """Generate an embedding for a single query using Ollama.

    Creates a vector embedding for a query string using the Ollama API.

    Args:
        config: Configuration for the Ollama embedding service.
        query: The query string to embed.

    Returns:
        A vector embedding (sequence of floats) for the query.

    Raises:
        RuntimeError: If the number of embeddings returned is not exactly 1.
    """
    client = AsyncClient(host=config.base_url)  # TODO: cache this?
    response = await client.embed(config.model, query)
    if len(response.embeddings) != 1:
        raise RuntimeError(f"expected 1 embedding but got {len(response.embeddings)}")
    return response.embeddings[0]



from collections.abc import AsyncGenerator, Callable
from typing import Any, Literal

from pydantic import BaseModel
from typing_extensions import override

from ..embeddings import (
    ApiKeyMixin,
    Embedder,
    EmbeddingResponse,
    EmbeddingVector,
    Usage,
    logger,
)
from .voyageai import voyage_max_tokens_per_batch, voyage_token_counter


class LiteLLM(ApiKeyMixin, BaseModel, Embedder):
    """
    Embedder that uses LiteLLM to embed documents into vector representations.

    Attributes:
        implementation (Literal["litellm"]): The literal identifier for this
            implementation.
        model (str): The name of the embedding model.
        api_key_name (str): The API key name.
        extra_options (dict): Additional litellm-specific options
    """

    implementation: Literal["litellm"]
    model: str
    extra_options: dict[str, Any] = {}

    @override
    async def embed(
        self, documents: list[str]
    ) -> AsyncGenerator[list[EmbeddingVector], None]:
        """
        Embeds a list of documents into vectors using LiteLLM.

        Args:
            documents (list[str]): A list of documents to be embedded.

        Returns:
            Sequence[EmbeddingVector]: The embeddings for each document.
        """
        await logger.adebug(f"Chunks produced: {len(documents)}")
        token_counter = self._token_counter()
        logger.debug("counting tokens")
        chunk_lengths = (
            [0 for _ in documents]
            if token_counter is None
            else [token_counter(doc) for doc in documents]
        )
        logger.debug("batching")
        async for embeddings in self.batch_chunks_and_embed(documents, chunk_lengths):
            yield embeddings

    @override
    def _max_chunks_per_batch(self) -> int:
        # Note: deferred import to avoid import overhead
        import litellm

        _, custom_llm_provider, _, _ = litellm.get_llm_provider(self.model)  # type: ignore
        match custom_llm_provider:
            case "cohere":
                return 96  # see https://docs.cohere.com/v1/reference/embed#request.body.texts
            case "openai":
                return 2048  # see https://platform.openai.com/docs/api-reference/embeddings/create
            case "azure":
                return 2048  # https://learn.microsoft.com/en-us/azure/ai-services/openai/how-to/embeddings?tabs=console#verify-inputs-dont-exceed-the-maximum-length
            case "bedrock":
                return 96  # NOTE: currently (Jan 2025) Bedrock only supports embeddings with Cohere or Titan models. The Titan API only processes one input per request, which LiteLLM already handles under the hood. We assume that the Cohere API has the same input limits as above.
            case "huggingface":
                return 2048  # NOTE: There is not documented limit. In testing we got a response for a request with 10k (short) inputs.
            case "mistral":
                return 128  # NOTE: this is chosen somewhat arbitrarily.
            case "vertex_ai":
                return 250  # FIXME: this applies to `us-central-1` only (otherwise 5). See https://cloud.google.com/vertex-ai/generative-ai/docs/embeddings/get-text-embeddings#get_text_embeddings_for_a_snippet_of_text
            case "voyage":
                return 128  # see https://docs.voyageai.com/reference/embeddings-api
            case _:
                logger.warn(
                    f"unknown provider '{custom_llm_provider}', falling back to conservative max chunks per batch"
                )
                return 5

    @override
    def _max_tokens_per_batch(self) -> int | None:
        # Note: deferred import to avoid import overhead
        import litellm

        model, custom_llm_provider, _, _ = litellm.get_llm_provider(self.model)  # type: ignore
        match custom_llm_provider:
            case "mistral":
                return 16_000  # No official documentation, see: https://github.com/langchain-ai/langchain/blob/33354f984fba660e71ca0039cfbd3cf37643cfab/libs/partners/mistralai/langchain_mistralai/embeddings.py#L25
            case "vertex_ai":
                return 20_000  # See https://cloud.google.com/vertex-ai/generative-ai/docs/embeddings/get-text-embeddings#get_text_embeddings_for_a_snippet_of_text
            case "openai" | "azure":
                return 600_000
            case "voyage":
                return voyage_max_tokens_per_batch(model)
            case _:
                return None

    def _token_counter(self) -> Callable[[str], int] | None:
        # Note: deferred import to avoid import overhead
        import litellm

        model, custom_llm_provider, _, _ = litellm.get_llm_provider(self.model)  # type: ignore
        match custom_llm_provider:
            case "mistral":
                # Note: deferred import to avoid import overhead
                from mistral_common.tokens.tokenizers.mistral import MistralTokenizer

                m_tokenizer = MistralTokenizer.from_model(model, strict=True)  # type: ignore

                def token_counter(text: str) -> int:
                    return len(
                        m_tokenizer.instruct_tokenizer.tokenizer.encode(  # type: ignore
                            text, False, False
                        )
                    )

                return token_counter
            case "vertex_ai":
                # Note: deferred import to avoid import overhead
                from vertexai.language_models import TextEmbeddingModel  # type: ignore

                v_tokenizer = TextEmbeddingModel.from_pretrained(model)

                def token_counter(text: str) -> int:
                    # NOTE: This is hideously inefficient, as evey call to
                    # count_tokens makes an API request to the CountTokens API
                    return v_tokenizer.count_tokens([text]).total_tokens

                return token_counter
            case "openai" | "azure":
                try:
                    # Note: deferred import to avoid import overhead
                    import tiktoken

                    encoder = tiktoken.encoding_for_model(model)
                except KeyError:
                    logger.warning(f"Tokenizer for the model {self.model} not found.")
                    encoder = None

                def token_counter(text: str) -> int:
                    if encoder is None:
                        return 0
                    return len(encoder.encode(text))

                return token_counter
            case "voyage":
                return voyage_token_counter(model, self._api_key)
            case _:
                return None

    @override
    async def call_embed_api(self, documents: list[str]) -> EmbeddingResponse:
        # Note: deferred import to avoid import overhead
        import litellm

        # Without `suppress_debug_info`, LiteLLM writes the following into stdout:
        # Provider List: https://docs.litellm.ai/docs/providers
        # This is useless, and confusing, so we suppress it.
        litellm.suppress_debug_info = True
        api_key = None if self.api_key_name is None else self._api_key
        response = await litellm.aembedding(  # type: ignore
            model=self.model,
            input=documents,
            api_key=api_key,
            **self.extra_options,
        )
        usage = (
            Usage(
                prompt_tokens=response.usage.prompt_tokens,
                total_tokens=response.usage.total_tokens,
            )
            if response.usage is not None
            else Usage(prompt_tokens=0, total_tokens=0)
        )
        return EmbeddingResponse(
            embeddings=[d["embedding"] for d in response["data"]], usage=usage
        )



import dotenv
import psycopg
import pytest
from psycopg.errors import Diagnostic

dotenv.load_dotenv()


def does_test_user_exist(cur: psycopg.Cursor) -> bool:
    cur.execute("""
        select count(*) > 0
        from pg_catalog.pg_roles
        where rolname = 'test'
    """)
    res = cur.fetchone()
    assert res is not None
    return res[0]


def create_test_user(cur: psycopg.Cursor) -> None:
    if not does_test_user_exist(cur):
        cur.execute("create user test password 'test'")


def does_test_db_exist(cur: psycopg.Cursor) -> bool:
    cur.execute("""
        select count(*) > 0
        from pg_catalog.pg_database
        where datname = 'test'
    """)
    res = cur.fetchone()
    assert res is not None
    return res[0]


def drop_test_db(cur: psycopg.Cursor) -> None:
    cur.execute(
        "select pg_terminate_backend(pid) from pg_stat_activity where datname = 'test'"
    )
    cur.execute("drop database test with (force)")


def create_test_db(cur: psycopg.Cursor) -> None:
    if does_test_db_exist(cur):
        drop_test_db(cur)
    cur.execute("create database test owner test")


@pytest.fixture(autouse=True)
def set_up_test_db() -> None:
    # create a test user and test database owned by the test user
    with psycopg.connect(
        "postgres://postgres@127.0.0.1:5432/postgres", autocommit=True
    ) as con:
        with con.cursor() as cur:
            create_test_user(cur)
            create_test_db(cur)
    # grant some things to the test user in the test database
    with psycopg.connect(
        "postgres://postgres@127.0.0.1:5432/test", autocommit=True
    ) as con:
        with con.cursor() as cur:
            cur.execute("grant execute on function pg_read_binary_file(text) to test")
            cur.execute("grant pg_read_server_files to test")
    # use the test user to create the extension in the test database
    import pgai

    pgai.install("postgres://test@127.0.0.1:5432/test")


def detailed_notice_handler(diag: Diagnostic) -> None:
    print(f"""
    Severity: {diag.severity}
    Message:  {diag.message_primary}
    Detail:   {diag.message_detail}
    Hint:     {diag.message_hint}
    """)



import asyncio

import psycopg
import pytest
from psycopg.rows import namedtuple_row
from psycopg.sql import SQL, Identifier
from testcontainers.postgres import PostgresContainer  # type: ignore

import pgai
from pgai.vectorizer import Vectorizer, Worker
from pgai.vectorizer.features import Features
from pgai.vectorizer.worker_tracking import WorkerTracking

from .conftest import create_connection_url


def create_database(dbname: str, postgres_container: PostgresContainer) -> None:
    with (
        psycopg.connect(
            postgres_container.get_connection_url(), autocommit=True
        ) as con,
        con.cursor() as cur,
    ):
        cur.execute(
            SQL("drop database if exists {dbname} with (force)").format(
                dbname=Identifier(dbname)
            )
        )
        cur.execute(SQL("create database {dbname}").format(dbname=Identifier(dbname)))


async def _vectorizer_test_after_install(
    postgres_container: PostgresContainer,
    dbname: str,
    ai_extension_features: bool = False,
):
    db_url = create_connection_url(postgres_container, dbname=dbname)
    with (
        psycopg.connect(db_url, autocommit=True, row_factory=namedtuple_row) as con,
        con.cursor() as cur,
    ):
        if ai_extension_features:
            cur.execute("create extension if not exists ai cascade")
        cur.execute("drop table if exists note0")
        cur.execute("""
                create table note0
                ( id bigint not null primary key generated always as identity
                , note text not null
                )
            """)
        # insert 5 rows into source
        cur.execute("""
                insert into note0 (note)
                select 'how much wood would a woodchuck chuck if a woodchuck could chuck wood'
                from generate_series(1, 5)
            """)  # noqa
        # insert 5 rows into source
        cur.execute("""
                insert into note0 (note)
                select 'if a woodchuck could chuck wood, a woodchuck would chuck as much wood as he could'
                from generate_series(1, 5)
            """)  # noqa
        # create a vectorizer for the table
        additional_args = ""
        if ai_extension_features:
            additional_args = """
                , scheduling=>
                    ai.scheduling_timescaledb
                    ( interval '5m'
                    , initial_start=>'2050-01-06'::timestamptz
                    , timezone=>'America/Chicago'
                    )
                , indexing=>ai.indexing_diskann(min_rows=>10)
            """

        cur.execute(f"""
                select ai.create_vectorizer
                ( 'note0'::regclass
                , loading=>ai.loading_column('note')
                , embedding=>ai.embedding_openai('text-embedding-3-small', 3)
                , formatting=>ai.formatting_python_template('$id: $chunk')
                , chunking=>ai.chunking_character_text_splitter()
                {additional_args}
                , grant_to=>null
                , enqueue_existing=>true
                )
            """)
        row = cur.fetchone()
        if row is None:
            raise ValueError("vectorizer_id is None")
        vectorizer_id = row[0]
        if not isinstance(vectorizer_id, int):
            raise ValueError("vectorizer_id is not an integer")

        cur.execute("select * from ai.vectorizer where id = %s", (vectorizer_id,))
        vectorizer_expected = cur.fetchone()

        worker = Worker(db_url)

        # test cli.get_vectorizer_ids
        assert len(worker._get_vectorizer_ids()) == 1  # type: ignore
        assert len(worker._get_vectorizer_ids([42, 19])) == 0  # type: ignore
        assert len(worker._get_vectorizer_ids([vectorizer_id, 19])) == 1  # type: ignore
        assert len(worker._get_vectorizer_ids([vectorizer_id])) == 1  # type: ignore

        # test cli.get_vectorizer
        features = Features.for_testing_latest_version()
        vectorizer_actual: Vectorizer = worker._get_vectorizer(  # type: ignore
            vectorizer_id, features
        )
        assert vectorizer_actual is not None
        assert vectorizer_expected.source_table == vectorizer_actual.source_table  # type: ignore

        # run the vectorizer
        worker_tracking = WorkerTracking(db_url, 500, features, "0.0.1")

        await vectorizer_actual.run(db_url, features, worker_tracking, 1)

        # make sure the queue was emptied
        cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer_id,))
        actual = cur.fetchone()[0]  # type: ignore
        assert actual == 0

        # make sure we got 10 rows out
        cur.execute(
            SQL("select count(*) from {target_schema}.{target_table}").format(
                target_schema=Identifier(
                    vectorizer_expected.config["destination"]["target_schema"]  # type: ignore
                ),  # type: ignore
                target_table=Identifier(
                    vectorizer_expected.config["destination"]["target_table"]  # type: ignore
                ),  # type: ignore
            )
        )
        actual = cur.fetchone()[0]  # type: ignore
        assert actual == 10

        # make sure the chunks were formatted correctly
        cur.execute(
            SQL("""
                select count(*) = count(*)
                filter (where chunk = format('%s: %s', id, note))
                from {view_schema}.{view_name}
                """).format(
                view_schema=Identifier(
                    vectorizer_expected.config["destination"]["view_schema"]  # type: ignore
                ),  # type: ignore
                view_name=Identifier(
                    vectorizer_expected.config["destination"]["view_name"]  # type: ignore
                ),  # type: ignore
            )
        )
        actual = cur.fetchone()[0]  # type: ignore
        assert actual is True


@pytest.mark.asyncio
async def test_vectorizer_internal(postgres_container: PostgresContainer):
    db = "vcli0"
    create_database(db, postgres_container)
    _db_url = create_connection_url(postgres_container, dbname=db)
    with (
        psycopg.connect(_db_url, autocommit=True, row_factory=namedtuple_row) as con,
        con.cursor() as cur,
    ):
        cur.execute("create extension if not exists vectorscale cascade")
        pgai.install(_db_url)
        cur.execute("create extension if not exists timescaledb")
    await _vectorizer_test_after_install(postgres_container, db)


@pytest.mark.asyncio
async def test_vectorizer_weird_pk(postgres_container: PostgresContainer):
    # make sure we can handle a multi-column primary key with "interesting" data types
    # this has implications on the COPY with binary format logic in the vectorizer
    db = "vcli1"
    create_database(db, postgres_container)
    db_url = postgres_container.get_connection_url()
    with (
        psycopg.connect(db_url, autocommit=True, row_factory=namedtuple_row) as con,
        con.cursor() as cur,
    ):
        cur.execute("create extension if not exists vectorscale cascade")
        cur.execute("create extension if not exists timescaledb")
        pgai.install(db_url)
        cur.execute("drop table if exists weird")
        cur.execute("""
                create table weird
                ( a text[] not null
                , b varchar(3) not null
                , c timestamp with time zone not null
                , d tstzrange not null
                , note text not null
                -- use a different ordering for pk to ensure we handle it
                , primary key (a, c, b, d)
                )
            """)
        # create a vectorizer for the table
        cur.execute("""
                select ai.create_vectorizer
                ( 'weird'::regclass
                , loading=>ai.loading_column('note')
                , embedding=>ai.embedding_openai('text-embedding-3-small', 3)
                , formatting=>ai.formatting_python_template('$chunk')
                , chunking=>ai.chunking_character_text_splitter()
                , grant_to=>null
                , enqueue_existing=>true
                )
            """)
        row = cur.fetchone()
        if row is None:
            raise ValueError("vectorizer_id is None")
        vectorizer_id = row[0]
        if not isinstance(vectorizer_id, int):
            raise ValueError("vectorizer_id is not an integer")

        cur.execute("select * from ai.vectorizer where id = %s", (vectorizer_id,))
        vectorizer_expected = cur.fetchone()

        # test worker._get_vectorizer
        worker = Worker(db_url)
        features = Features.for_testing_latest_version()
        vectorizer_actual: Vectorizer = worker._get_vectorizer(  # type: ignore
            vectorizer_id, features
        )
        assert vectorizer_actual is not None
        assert vectorizer_expected.source_table == vectorizer_actual.source_table  # type: ignore

        # insert 7 rows into source
        cur.execute("""
                insert into weird (a, b, c, d, note)
                select
                  array['larry', 'moe', 'curly']
                , 'xyz'
                , t
                , tstzrange(t, t + interval '1d', '[)')
                , 'if two witches watch two watches, which witch watches which watch'
                from generate_series('2025-01-06'::timestamptz, '2025-01-12'::timestamptz, interval '1d') t
            """)  # noqa

        # run the vectorizer
        features = Features.for_testing_latest_version()
        worker_tracking = WorkerTracking(db_url, 500, features, "0.0.1")
        await vectorizer_actual.run(db_url, features, worker_tracking, 1)

        # make sure the queue was emptied
        cur.execute("select ai.vectorizer_queue_pending(%s)", (vectorizer_id,))
        actual = cur.fetchone()[0]  # type: ignore
        assert actual == 0

        # make sure we got 7 rows out
        cur.execute(
            SQL("select count(*) from {target_schema}.{target_table}").format(
                target_schema=Identifier(
                    vectorizer_expected.config["destination"]["target_schema"]  # type: ignore
                ),  # type: ignore
                target_table=Identifier(
                    vectorizer_expected.config["destination"]["target_table"]  # type: ignore
                ),  # type: ignore
            )
        )
        actual = cur.fetchone()[0]  # type: ignore
        assert actual == 7


@pytest.mark.asyncio
@pytest.mark.parametrize("async_install", [True, False])
async def test_vectorizer_install_twice(
    postgres_container: PostgresContainer, async_install: bool
):
    db = "ainstall2"
    create_database(db, postgres_container)
    _db_url = create_connection_url(postgres_container, dbname=db)
    if async_install:
        await pgai.ainstall(_db_url)
        await pgai.ainstall(_db_url)
    else:
        pgai.install(_db_url)
        pgai.install(_db_url)

    with pytest.raises(psycopg.errors.DuplicateObject):
        if async_install:
            await pgai.ainstall(_db_url, strict=True)
        else:
            pgai.install(_db_url, strict=True)

    # test the vectorizer
    with (
        psycopg.connect(_db_url, autocommit=True, row_factory=namedtuple_row) as con,
        con.cursor() as cur,
    ):
        cur.execute("create extension if not exists timescaledb")
    await _vectorizer_test_after_install(postgres_container, db)


@pytest.mark.postgres_params(set_executor_url=True)
@pytest.mark.asyncio
@pytest.mark.parametrize("async_install", [True, False])
async def test_vectorizer_install_need_ai_extension(
    postgres_container: PostgresContainer, async_install: bool
):
    # the pytest mark set the ai.external_functions_executor_url to http://www.example.com

    db = "need_ai_extension"
    create_database(db, postgres_container)
    _db_url = create_connection_url(postgres_container, dbname=db)
    if async_install:
        await pgai.ainstall(_db_url)
    else:
        pgai.install(_db_url)

    with (
        psycopg.connect(_db_url, autocommit=True, row_factory=namedtuple_row) as con,
        con.cursor() as cur,
    ):
        cur.execute("select extname from pg_extension where extname = 'ai'")
        result = cur.fetchone()
        assert result is not None


@pytest.mark.asyncio
@pytest.mark.parametrize("async_install", [True, False])
async def test_vectorizer_install_no_ai_extension(
    postgres_container: PostgresContainer, async_install: bool
):
    # by default, the ai extension should not be installed
    db = "no_ai_extension"
    create_database(db, postgres_container)
    _db_url = create_connection_url(postgres_container, dbname=db)
    if async_install:
        await pgai.ainstall(_db_url)
    else:
        pgai.install(_db_url)

    with (
        psycopg.connect(_db_url, autocommit=True, row_factory=namedtuple_row) as con,
        con.cursor() as cur,
    ):
        cur.execute("select extname from pg_extension where extname = 'ai'")
        result = cur.fetchone()
        assert result is None


@pytest.mark.asyncio
@pytest.mark.parametrize("async_install", [True, False])
async def test_vectorizer_install_old_ai_extension(
    postgres_container: PostgresContainer, async_install: bool
):
    db = "old_ai_extension"
    create_database(db, postgres_container)
    _db_url = create_connection_url(postgres_container, dbname=db)

    with (
        psycopg.connect(_db_url, autocommit=True, row_factory=namedtuple_row) as con,
        con.cursor() as cur,
    ):
        cur.execute("create extension if not exists ai version '0.9.0' cascade")

    with pytest.raises(
        Exception, match="The ai extension is outdated and must be upgraded"
    ):
        if async_install:
            await pgai.ainstall(_db_url)
        else:
            pgai.install(_db_url)


@pytest.mark.asyncio
@pytest.mark.parametrize("async_install", [True, False])
async def test_vectorizer_install_vector_in_different_schema(
    postgres_container: PostgresContainer, async_install: bool
):
    db = "vector_in_different_schema"
    create_database(db, postgres_container)
    _db_url = create_connection_url(postgres_container, dbname=db)

    with (
        psycopg.connect(_db_url, autocommit=True, row_factory=namedtuple_row) as con,
        con.cursor() as cur,
    ):
        cur.execute("create schema other")
        cur.execute("create extension if not exists vector schema other")
        cur.execute("create extension if not exists timescaledb")
        cur.execute(f"alter database {db} set search_path = public,other")

    if async_install:
        await pgai.ainstall(_db_url)
    else:
        pgai.install(_db_url)

    await _vectorizer_test_after_install(postgres_container, db)


async def _vectorizer_setup_simplevectorizer(
    postgres_container: PostgresContainer,
    dbname: str,
):
    db_url = create_connection_url(postgres_container, dbname=dbname)
    with (
        psycopg.connect(db_url, autocommit=True, row_factory=namedtuple_row) as con,
        con.cursor() as cur,
    ):
        table_name = "notes_simple"
        cur.execute(f"drop table if exists {table_name}")
        cur.execute(f"""
                create table {table_name}
                ( id bigint not null primary key generated always as identity
                , note text not null
                )
        """)
        # insert 5 rows into source
        cur.execute(f"""
                insert into {table_name} (note)
                select 'how much wood would a woodchuck chuck if a woodchuck could chuck wood'
                from generate_series(1, 5)
            """)  # noqa
        # insert 5 rows into source
        cur.execute(f"""
                insert into {table_name} (note)
                select 'if a woodchuck could chuck wood, a woodchuck would chuck as much wood as he could'
                from generate_series(1, 5)
            """)  # noqa

        cur.execute(f"""
                select ai.create_vectorizer
                ( '{table_name}'::regclass
                , loading=>ai.loading_column('note')
                , embedding=>ai.embedding_openai('text-embedding-3-small', 3)
                , formatting=>ai.formatting_python_template('$id: $chunk')
                , chunking=>ai.chunking_character_text_splitter()
                , grant_to=>null
                , enqueue_existing=>true
                )
            """)


@pytest.mark.asyncio
async def test_vectorizer_run_once_with_shutdown(
    postgres_container: PostgresContainer,
):
    db = "run_once_with_shutdown"
    create_database(db, postgres_container)
    db_url = create_connection_url(postgres_container, dbname=db)
    await pgai.ainstall(db_url)

    await _vectorizer_setup_simplevectorizer(postgres_container, db)

    worker = Worker(db_url, once=True)
    task = asyncio.create_task(worker.run())
    await worker.request_graceful_shutdown()
    result = await asyncio.wait_for(task, timeout=300)
    assert result is None


@pytest.mark.asyncio
async def test_vectorizer_run_with_shutdown(
    postgres_container: PostgresContainer,
):
    db = "run_with_shutdown"
    create_database(db, postgres_container)
    db_url = create_connection_url(postgres_container, dbname=db)
    await pgai.ainstall(db_url)

    await _vectorizer_setup_simplevectorizer(postgres_container, db)

    worker = Worker(db_url)
    task = asyncio.create_task(worker.run())
    await worker.request_graceful_shutdown()
    result = await asyncio.wait_for(task, timeout=300)
    assert result is None



import os

import psycopg
import pytest

# skip tests in this module if disabled
enable_vectorizer_tests = os.getenv("ENABLE_VECTORIZER_TESTS")
if enable_vectorizer_tests == "0":
    pytest.skip(allow_module_level=True)


def db_url(user: str) -> str:
    return f"postgres://{user}@127.0.0.1:5432/test"


def test_parsing():
    tests = [
        (
            "select ai.parsing_docling()",
            {
                "config_type": "parsing",
                "implementation": "docling",
            },
        ),
        (
            "select ai.parsing_pymupdf()",
            {
                "config_type": "parsing",
                "implementation": "pymupdf",
            },
        ),
        (
            "select ai.parsing_auto()",
            {
                "config_type": "parsing",
                "implementation": "auto",
            },
        ),
        (
            "select ai.parsing_none()",
            {
                "config_type": "parsing",
                "implementation": "none",
            },
        ),
    ]

    with psycopg.connect(db_url("test")) as con:
        with con.cursor() as cur:
            for query, expected in tests:
                cur.execute(query)
                actual = cur.fetchone()[0]
                assert actual.keys() == expected.keys()
                for k, v in actual.items():
                    assert k in expected and v == expected[k]


def test_validate_parsing():
    ok = [
        """
        select ai._validate_parsing(
            ai.parsing_auto(),
            ai.loading_column('body'),
            'public',
            'thing'
        )
        """,
        """
        select ai._validate_parsing(
            ai.parsing_auto(),
            ai.loading_column('column'),
            'public',
            'thing'
        )
        """,
        """
        select ai._validate_parsing(
            ai.parsing_docling(),
            ai.loading_column('column'),
            'public',
            'thing'
        )
        """,
        """
        select ai._validate_parsing(
            ai.parsing_pymupdf(),
            ai.loading_column('column'),
            'public',
            'thing'
        )
        """,
    ]
    bad = [
        (
            """
            select ai._validate_parsing(
                ai.parsing_docling(),
                ai.loading_column('body'),
                'public',
                'thing'
            )
            """,
            "parsing_docling must be used with a bytea column",
        ),
        (
            """
            select ai._validate_parsing(
                ai.parsing_pymupdf(),
                ai.loading_column('body'),
                'public',
                'thing'
            )
            """,
            "parsing_pymupdf must be used with a bytea column",
        ),
        (
            """
            select ai._validate_parsing(
                ai.parsing_none(),
                ai.loading_column('document'),
                'public',
                'thing'
            )
            """,
            "cannot use parsing_none with bytea columns",
        ),
    ]
    with psycopg.connect(db_url("test"), autocommit=True) as con:
        with con.cursor() as cur:
            cur.execute("drop table if exists public.thing;")
            cur.execute(
                "create table public.thing (id int, color text, weight float, body text, document bytea)"
            )
            for query in ok:
                cur.execute(query)
                assert True
            for query, err in bad:
                try:
                    cur.execute(query)
                except psycopg.ProgrammingError as ex:
                    msg = str(ex.args[0])
                    assert len(msg) >= len(err) and msg[: len(err)] == err
                else:
                    pytest.fail(f"expected exception: {err}")



import pytest
from psycopg import Connection
from psycopg.rows import dict_row

from tests.vectorizer.cli.conftest import (
    TestDatabase,
    configure_vectorizer,
    run_vectorizer_worker,
    setup_source_table,
)


@pytest.mark.parametrize(
    "column_def",
    [
        "new_column text",
        "new_column text NOT NULL DEFAULT 'default_value'",
    ],
)
def test_additional_columns_are_added_to_target_table(
    cli_db: tuple[TestDatabase, Connection],
    cli_db_url: str,
    column_def: str,
):
    """Test that if additional columns are added to the target table,
    the vectorizer still works"""
    _, connection = cli_db
    table_name = setup_source_table(connection, 2)
    vectorizer_id = configure_vectorizer(
        table_name,
        cli_db[1],
    )
    with connection.cursor(row_factory=dict_row) as cur:
        cur.execute(f"ALTER TABLE blog_embedding_store ADD COLUMN {column_def}")  # type: ignore

    result = run_vectorizer_worker(cli_db_url, vectorizer_id)
    print(result.stdout)
    assert result.exit_code == 0

    with connection.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT * FROM blog_embedding_store")
        rows = cur.fetchall()
        assert len(rows) == 2


def test_embedding_column_removal_and_readd(
    cli_db: tuple[TestDatabase, Connection],
    cli_db_url: str,
):
    """Test that the vectorizer still works when the embedding column is removed,
    another column is added, and then the embedding column is re-added."""
    _, connection = cli_db
    table_name = setup_source_table(connection, 2)
    vectorizer_id = configure_vectorizer(
        table_name,
        cli_db[1],
    )

    # First run to create original rows
    result = run_vectorizer_worker(cli_db_url, vectorizer_id)
    assert result.exit_code == 0

    # Check original rows were created
    with connection.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT * FROM blog_embedding_store")
        rows = cur.fetchall()
        assert len(rows) == 2
        # Verify embedding column exists
        assert "embedding" in rows[0]

        # Drop View so we can change column order
        cur.execute("DROP VIEW IF EXISTS blog_embedding")

        # Remove embedding column
        cur.execute("ALTER TABLE blog_embedding_store DROP COLUMN embedding")

        # Add another optional column
        cur.execute("ALTER TABLE blog_embedding_store ADD COLUMN extra_data text")

        # Re-add embedding column with same type
        cur.execute(
            "ALTER TABLE blog_embedding_store ADD COLUMN embedding vector(1536)"
        )

        # Remove original rows
        cur.execute("DELETE FROM blog")

        # Add new rows
        values = [(i, i, f"post_{i}") for i in range(1, 3)]
        cur.executemany(
            "INSERT INTO blog(id, id2, content) VALUES (%s, %s, %s)",
            values,
        )

    # Run vectorizer again
    result = run_vectorizer_worker(cli_db_url, vectorizer_id)
    print(result.stdout)
    assert result.exit_code == 0

    # Verify vectorizer still works
    with connection.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT * FROM blog_embedding_store")
        rows = cur.fetchall()
        assert len(rows) == 2
        # Verify embedding column exists and has data
        assert "embedding" in rows[0]
        assert rows[0]["embedding"] is not None



import os

import psycopg
import pytest

# skip tests in this module if disabled
enable_vectorizer_tests = os.getenv("ENABLE_VECTORIZER_TESTS")
if enable_vectorizer_tests == "0":
    pytest.skip(allow_module_level=True)


def db_url(user: str) -> str:
    return f"postgres://{user}@127.0.0.1:5432/test"


def test_scheduling_none():
    tests = [
        (
            "select ai.scheduling_none()",
            {
                "implementation": "none",
                "config_type": "scheduling",
            },
        ),
    ]
    with psycopg.connect(db_url("test")) as con:
        with con.cursor() as cur:
            for query, expected in tests:
                cur.execute(query)
                actual = cur.fetchone()[0]
                assert actual.keys() == expected.keys()
                for k, v in actual.items():
                    assert k in expected and v == expected[k]


def test_scheduling_timescaledb():
    tests = [
        (
            "select ai.scheduling_timescaledb()",
            {
                "implementation": "timescaledb",
                "config_type": "scheduling",
                "schedule_interval": "00:05:00",
            },
        ),
        (
            "select ai.scheduling_timescaledb(interval '10m')",
            {
                "implementation": "timescaledb",
                "config_type": "scheduling",
                "schedule_interval": "00:10:00",
            },
        ),
        (
            "select ai.scheduling_timescaledb(interval '1h', timezone=>'America/Chicago')",
            {
                "implementation": "timescaledb",
                "config_type": "scheduling",
                "schedule_interval": "01:00:00",
                "timezone": "America/Chicago",
            },
        ),
        (
            "select ai.scheduling_timescaledb(interval '10m', fixed_schedule=>true, timezone=>'America/Chicago')",
            {
                "implementation": "timescaledb",
                "config_type": "scheduling",
                "schedule_interval": "00:10:00",
                "timezone": "America/Chicago",
                "fixed_schedule": True,
            },
        ),
        (
            """
            select ai.scheduling_timescaledb
            ( interval '15m'
            , initial_start=>'2025-01-06 America/Chicago'::timestamptz
            , fixed_schedule=>false
            , timezone=>'America/Chicago'
            )
            """,
            {
                "implementation": "timescaledb",
                "config_type": "scheduling",
                "schedule_interval": "00:15:00",
                "timezone": "America/Chicago",
                "fixed_schedule": False,
                "initial_start": "2025-01-06T06:00:00+00:00",
            },
        ),
    ]
    with psycopg.connect(db_url("test")) as con:
        with con.cursor() as cur:
            for query, expected in tests:
                cur.execute(query)
                actual = cur.fetchone()[0]
                assert actual.keys() == expected.keys()
                for k, v in actual.items():
                    assert k in expected and v == expected[k]


@pytest.mark.parametrize(
    "setting,expected",
    [
        (
            None,  # default case - no setting
            {
                "implementation": "none",
                "config_type": "scheduling",
            },
        ),
        (
            "scheduling_timescaledb",
            {
                "implementation": "timescaledb",
                "config_type": "scheduling",
                "schedule_interval": "00:05:00",
            },
        ),
        (
            "scheduling_none",
            {
                "implementation": "none",
                "config_type": "scheduling",
            },
        ),
        (
            "invalid_value",  # should default to none
            {
                "implementation": "none",
                "config_type": "scheduling",
            },
        ),
    ],
)
def test_scheduling_default(setting, expected):
    with psycopg.connect(db_url("test")) as con:
        with con.cursor() as cur:
            if setting is not None:
                cur.execute(f"set ai.scheduling_default = '{setting}'")
            cur.execute("select ai._resolve_scheduling_default()")
            actual = cur.fetchone()[0]
            assert actual.keys() == expected.keys()
            for k, v in actual.items():
                assert k in expected and v == expected[k]


def test_validate_scheduling():
    ok = [
        "select ai._validate_scheduling(ai.scheduling_none())",
        "select ai._validate_scheduling(ai.scheduling_timescaledb())",
    ]
    bad = [
        (
            "select ai._validate_scheduling(ai.indexing_hnsw(opclass=>'peter'))",
            "invalid config_type for scheduling config",
        ),
        (
            """select ai._validate_scheduling('{"config_type": "scheduling"}'::jsonb)""",
            "scheduling implementation not specified",
        ),
        (
            """
            select ai._validate_scheduling
            ( '{"config_type": "scheduling", "implementation": "grandfather clock"}'::jsonb
            )
            """,
            'unrecognized scheduling implementation: "grandfather clock"',
        ),
    ]
    with psycopg.connect(db_url("test"), autocommit=True) as con:
        with con.cursor() as cur:
            for query in ok:
                cur.execute(query)
                assert True
            for query, err in bad:
                try:
                    cur.execute(query)
                except psycopg.ProgrammingError as ex:
                    msg = str(ex.args[0])
                    assert len(msg) >= len(err) and msg[: len(err)] == err
                else:
                    pytest.fail(f"expected exception: {err}")



import asyncio
from collections.abc import Sequence
from pathlib import Path
from typing import Any

import click
import vcr  # type: ignore
from vcr.record_mode import RecordMode  # type: ignore

from pgai import __version__ as pgai_version
from pgai.cli import TimeDurationParamType, async_run_vectorizer_worker


@click.command(name="worker-with-vcr")
@click.version_option(version=pgai_version)
@click.option(
    "--cassette",
    type=click.STRING,
    default="wiki_openai_500",
    show_default=True,
    help="VCR cassette to use on the benchmark",
)
@click.option(
    "--record-mode",
    type=click.STRING,
    default="once",
    show_default=True,
    help="VCR record mode (all, any, new_episodes, none, once)",
)
@click.option(
    "-d",
    "--db-url",
    type=click.STRING,
    default="postgres://postgres@localhost:5432/postgres",
    show_default=True,
    help="The database URL to connect to",
)
@click.option(
    "-i",
    "--vectorizer-id",
    "vectorizer_ids",
    type=click.INT,
    multiple=True,
    help="Only fetch work from the given vectorizer ids. If not provided, all vectorizers will be fetched.",  # noqa
    default=[],
)
@click.option(
    "--log-level",
    type=click.Choice(
        ["DEBUG", "INFO", "WARN", "ERROR", "FATAL", "CRITICAL"], case_sensitive=False
    ),
    default="INFO",
)
@click.option(
    "--poll-interval",
    type=TimeDurationParamType(),
    default="5m",
    show_default=True,
    help="The interval, in duration string or integer (seconds), "
    "to wait before checking for new work after processing "
    "all available work in the queue.",
    # noqa
)
@click.option(
    "--once",
    type=click.BOOL,
    is_flag=True,
    default=False,
    show_default=True,
    help="Exit after processing all available work (implies --exit-on-error).",
)
@click.option(
    "-c",
    "--concurrency",
    type=click.IntRange(1),
    default=None,
    show_default=True,
)
@click.option(
    "--exit-on-error",
    type=click.BOOL,
    default=None,
    show_default=True,
    help="Exit immediately when an error occurs.",
)
def worker_with_vcr(
    cassette: str,
    record_mode: str,
    db_url: str,
    vectorizer_ids: Sequence[int],
    concurrency: int,
    log_level: str,
    poll_interval: int,
    once: bool,
    exit_on_error: bool | None,
) -> None:
    cassette_library_dir = Path(__file__).parent.joinpath("cassettes")
    cassette_library_dir.mkdir(exist_ok=True)

    def remove_set_cookie_header(response: dict[str, Any]):
        headers = response["headers"]
        headers_to_remove = ["set-cookie", "Set-Cookie"]
        for header in headers_to_remove:
            if header in headers:
                del headers[header]
        return response

    vcr_ = vcr.VCR(
        serializer="yaml",
        cassette_library_dir=str(cassette_library_dir),
        record_mode=RecordMode(record_mode.lower()),
        filter_headers=["authorization", "api-key"],
        match_on=["method", "scheme", "host", "port", "path", "query", "body"],
        before_record_response=remove_set_cookie_header,
    )
    with vcr_.use_cassette(cassette):  # type: ignore
        asyncio.run(
            async_run_vectorizer_worker(
                db_url,
                vectorizer_ids,
                log_level,
                poll_interval,
                once,
                concurrency,
                exit_on_error,
            )
        )


@click.group()
@click.version_option(version=pgai_version)
def vectorizer():
    pass


@click.group()
@click.version_option(version=pgai_version)
def cli():
    pass


vectorizer.add_command(worker_with_vcr)
cli.add_command(vectorizer)



import logging

import psycopg
from psycopg.rows import dict_row
from psycopg.sql import SQL, Identifier

from pgai.semantic_catalog.models import (
    ObjectDescription,
    Procedure,
    Table,
    View,
)
from pgai.semantic_catalog.sample import sample_table, sample_view

logger = logging.getLogger(__name__)


async def load_tables(
    con: psycopg.AsyncConnection, oids: list[int], sample_size: int = 0
) -> list[Table]:
    """Load table definitions from the database based on object IDs.

    Retrieves table metadata including columns, constraints, and indexes for the
    specified table OIDs. If sample_size is greater than 0, it also retrieves sample
    data from each table.

    Args:
        con: Asynchronous database connection object.
        oids: List of table object IDs to load.
        sample_size: Number of sample rows to retrieve from each table (default: 0).
            If 0, no sample data is retrieved.

    Returns:
        A list of Table objects with metadata and optionally sample data.

    Raises:
        AssertionError: If the list of oids is empty.
    """
    # TODO: add support for hypertable info
    # TODO: add support for partitioning info
    # TODO: add support for inheritance info
    # TODO: add support for foreign tables
    assert len(oids) > 0, "list of oids must not be empty"
    async with con.cursor(row_factory=dict_row) as cur:
        await cur.execute("select 1 from pg_extension where extname = 'timescaledb'")
        is_timescale = await cur.fetchone() is not None
        timescale_cte = """
        select
            null as schema_name
            , null as table_name, null as dimensions
        """
        if is_timescale:
            timescale_cte = """
            select
                h.schema_name
                , h.table_name
                , d.dimensions
            from
                _timescaledb_catalog.hypertable AS h
            join (
                select
                    d.hypertable_id
                    , json_agg(json_build_object(
                        'dimension_builder', d.dimension_builder
                        , 'column_name', d.column_name
                        , 'partition_func', d.partition_func
                        , 'partition_interval', d.partition_interval
                        , 'number_partitions', d.number_partitions
                    )) as dimensions
                from (
                    select
                    d.hypertable_id,
                    case
                        when d.partitioning_func is null then 'by_range'
                        when d.partitioning_func = 'get_partition_hash' then 'by_hash'
                        when f.function_type IN ('int2', 'int4', 'int8') then 'by_hash'
                        else 'by_range'
                    end AS dimension_builder,
                    d.column_name AS column_name,
                    case
                        when d.partitioning_func is null then null
                        when d.partitioning_func = 'get_partition_hash' then null
                        else format(
                            '%%s.%%s'
                            , d.partitioning_func_schema
                            , d.partitioning_func
                        )
                    end as partition_func,
                    case
                        when d.interval_length is null then null
                        else 'INTERVAL '''
                            || justify_interval(
                                d.interval_length * interval '1 microsecond'
                            )
                            || ''''
                    end as partition_interval,
                    d.num_slices as number_partitions
                    from _timescaledb_catalog.dimension as d
                    left join
                    (
                        select
                        n.nspname as function_schema
                        , p.proname as function_name
                        , t.typname as function_type
                        from pg_proc p
                        join pg_namespace n on n.oid = p.pronamespace
                        join pg_type t on t.oid = p.prorettype
                    ) as f on (
                        f.function_schema = d.partitioning_func_schema
                        and f.function_name = d.partitioning_func
                    )
                    order by d.id
                ) as d
                group by d.hypertable_id
            ) as d on d.hypertable_id = h.id
            where h.schema_name != '_timescaledb_internal'
            """

        await cur.execute(
            f"""
            with x as
            (
                select
                  'pg_catalog.pg_class'::regclass::oid as classid
                , k.oid as objid
                , n.nspname as schema_name
                , k.relname as table_name
                , case k.relpersistence
                    when 'u' then 'unlogged'
                    when 't' then 'temporary'
                  end as persistence
                , ( -- columns
                    select jsonb_agg
                    (
                      jsonb_build_object
                      ( 'classid', 'pg_catalog.pg_class'::regclass::oid
                      , 'objid', k.oid
                      , 'objsubid', a.attnum
                      , 'name', a.attname
                      , 'type', pg_catalog.format_type(a.atttypid, a.atttypmod)
                      , 'is_not_null', a.attnotnull
                      , 'default_value', pg_get_expr(d.adbin, d.adrelid)
                      )
                      order by a.attnum
                    )
                    from pg_attribute a
                    left outer join pg_attrdef d
                        on (a.attrelid = d.adrelid and a.attnum = d.adnum)
                    where a.attrelid = k.oid
                    and a.attnum > 0
                    and not a.attisdropped
                  ) as columns
                , ( -- constraints
                    select jsonb_agg
                    (
                        jsonb_build_object
                        ( 'name', c.conname
                        , 'definition', pg_get_constraintdef(c.oid)
                        )
                        order by c.oid
                    )
                    from pg_constraint c
                    where c.conrelid = k.oid
                  ) as constraints
                , ( -- indexes
                    select jsonb_agg
                    (
                      jsonb_build_object
                      ( 'name', x.relname
                      , 'definition', pg_get_indexdef(i.indexrelid)
                      )
                      order by i.indexrelid
                    )
                    from pg_index i
                    inner join pg_class x on (i.indexrelid = x.oid)
                    where i.indrelid = k.oid
                    and i.indisprimary = false -- already represented in constraints
                  ) as indexes
                from pg_class k
                inner join pg_namespace n on (k.relnamespace = n.oid)
                where k.oid = any(%s::oid[])
                and k.relkind in ('r', 'p', 'f')
            ),
            y as
            (
                {timescale_cte}
            )
            select x.*, y.dimensions
            from x
            left join y on (
                x.schema_name = y.schema_name
                and x.table_name = y.table_name
            )
        """,
            (oids,),
        )
        tables: list[Table] = []
        for row in await cur.fetchall():
            tables.append(Table.model_validate(row))
        if len(tables) != len(oids):
            logger.warning(
                f"{len(oids)} oids were provided but only {len(tables)} tables were loaded"
            )
        if sample_size > 0:
            logger.debug(
                f"sampling {sample_size} rows from each of {len(tables)} tables"
            )
            for table in tables:
                logger.debug(f"sampling {table.schema_name}.{table.table_name}")
                table.sample = await sample_table(
                    con, table.schema_name, table.table_name, limit=sample_size
                )
        return tables


async def load_views(
    con: psycopg.AsyncConnection, oids: list[int], sample_size: int = 0
) -> list[View]:
    """Load view definitions from the database based on object IDs.

    Retrieves view metadata including columns and definition SQL for the specified
    view OIDs. Handles both regular views and materialized views. If TimescaleDB is
    installed, it also identifies continuous aggregates. If sample_size is greater
    than 0, it also retrieves sample data from each view.

    Args:
        con: Asynchronous database connection object.
        oids: List of view object IDs to load.
        sample_size: Number of sample rows to retrieve from each view (default: 0).
            If 0, no sample data is retrieved.

    Returns:
        A list of View objects with metadata and optionally sample data.

    Raises:
        AssertionError: If the list of oids is empty.
    """
    assert len(oids) > 0, "list of oids must not be empty"
    async with con.cursor(row_factory=dict_row) as cur:
        await cur.execute("select 1 from pg_extension where extname = 'timescaledb'")
        is_timescale = await cur.fetchone() is not None
        continuous_aggregate = "false"
        materialized = "k.relkind = 'm'"
        view_def = "pg_get_viewdef(k.oid, true)"
        ts_join = ""
        if is_timescale:
            continuous_aggregate = "ca.view_definition is not null"
            materialized = f"{continuous_aggregate} or {materialized}"
            view_def = f"coalesce(ca.view_definition, {view_def})"
            ts_join = """
            left join timescaledb_information.continuous_aggregates ca
                on (ca.view_schema = n.nspname and ca.view_name = k.relname)
            """

        await cur.execute(
            f"""\
            select
              'pg_catalog.pg_class'::regclass::oid as classid
            , k.oid as objid
            , n.nspname as schema_name
            , k.relname as view_name
            , {materialized} as is_materialized
            , {continuous_aggregate} as is_continuous_aggregate
            , {view_def} as definition
            , ( -- columns
                select jsonb_agg
                (
                  jsonb_build_object
                  ( 'classid', 'pg_catalog.pg_class'::regclass::oid
                  , 'objid', k.oid
                  , 'objsubid', a.attnum
                  , 'name', a.attname
                  , 'type', pg_catalog.format_type(a.atttypid, a.atttypmod)
                  , 'is_not_null', a.attnotnull
                  , 'default_value', null::text
                  )
                  order by a.attnum
                )
                from pg_attribute a
                left outer join pg_attrdef d
                    on (a.attrelid = d.adrelid and a.attnum = d.adnum)
                where a.attrelid = k.oid
                and a.attnum > 0
                and not a.attisdropped
              ) as columns
            from pg_class k
            inner join pg_namespace n on (k.relnamespace = n.oid)
            {ts_join}
            where k.oid = any(%s::oid[])
            and k.relkind in ('v', 'm')
        """,
            (oids,),
        )
        views: list[View] = []
        for row in await cur.fetchall():
            views.append(View.model_validate(row))
        if len(views) != len(oids):
            logger.warning(
                f"{len(oids)} oids were provided but only {len(views)} views were loaded"
            )
        if sample_size > 0:
            logger.debug(f"sampling {sample_size} rows from each of {len(views)} views")
            for view in views:
                logger.debug(f"sampling {view.schema_name}.{view.view_name}")
                view.sample = await sample_view(
                    con, view.schema_name, view.view_name, limit=sample_size
                )
        return views


async def load_procedures(
    con: psycopg.AsyncConnection, oids: list[int]
) -> list[Procedure]:
    """Load procedure definitions from the database based on object IDs.

    Retrieves metadata for procedures, functions, and aggregates for the specified OIDs,
    including their full definition SQL.

    Args:
        con: Asynchronous database connection object.
        oids: List of procedure object IDs to load.

    Returns:
        A list of Procedure objects with metadata.

    Raises:
        AssertionError: If the list of oids is empty.
    """
    assert len(oids) > 0, "list of oids must not be empty"
    async with con.cursor(row_factory=dict_row) as cur:
        await cur.execute(
            """\
            select
              'pg_catalog.pg_proc'::regclass::oid as classid
            , p.oid as objid
            , n.nspname as schema_name
            , p.proname as proc_name
            , case p.prokind
                when 'f' then 'function'
                when 'w' then 'function'
                when 'p' then 'procedure'
                when 'a' then 'aggregate'
              end as kind
            , pg_get_function_identity_arguments(p.oid) as identity_args
            , case
                when p.prokind = 'a' then
                  'CREATE OR REPLACE AGGREGATE FUNCTION '
                  || quote_ident(n.nspname) || '.' || quote_ident(p.proname)
                  || '(' || pg_get_function_identity_arguments(p.oid) ||  ') (' || E'\n'
                  || '  sfunc = ' || a.aggtransfn::text
                  || ',' || E'\n' || '  stype = ' || tt.typname::text
                  || case
                    when a.aggmtransfn != 0 then
                      ',' || E'\n' || '  msfunc = ' || a.aggmtransfn::text || ','
                      || E'\n' || '  mstype = ' || tm.typname::text
                    else ''
                  end
                  || case
                    when a.aggfinalfn != 0 then
                      ',' || E'\n' || '  finalfunc = ' || a.aggfinalfn::text
                      || case
                        when a.aggfinalextra is true then
                          ', ' || E'\n' || '  finalfunc_extra'
                        else ''
                      end
                      || ',' || E'\n' || '  finalfunc_modify = '
                      || case
                        when a.aggfinalmodify = 'r' then 'READ_ONLY'
                        when a.aggfinalmodify = 's' then 'SHAREABLE'
                        when a.aggfinalmodify = 'w' then 'READ_WRITE'
                      end
                    else ''
                  end
                  || case
                    when a.aggcombinefn != 0 then
                      ',' || E'\n' || '  combinefunc = ' || a.aggcombinefn::text
                    else ''
                  end
                  || case
                    when a.aggserialfn != 0 then
                      ',' || E'\n' || '  serialfunc = ' || a.aggserialfn::text
                    else ''
                  end
                  || case
                    when a.aggdeserialfn != 0 then
                      ',' || E'\n' || '  deserialfunc = ' || a.aggdeserialfn::text
                    else ''
                  end
                  || case
                    when a.aggmfinalfn != 0 then
                      ',' || E'\n' || '  mfinalfunc = ' || a.aggmfinalfn::text
                      || case
                        when a.aggmfinalextra is true then
                        ', ' || E'\n' || '  mfinalfunc_extra'
                        else ''
                      end
                      || ',' || E'\n' || '  mfinalfunc_modify = '
                      || case
                        when a.aggmfinalmodify = 'r' then 'READ_ONLY'
                        when a.aggmfinalmodify = 's' then 'SHAREABLE'
                        when a.aggmfinalmodify = 'w' then 'READ_WRITE'
                      end
                      else ''
                  end
                  || case
                    when a.aggminvtransfn != 0 then ',' || E'\n'
                      || '  minvfunc = ' || a.aggminvtransfn::text
                    else ''
                  end
                  || case
                    when a.agginitval is not null then ',' || E'\n'
                      || '  initcond = ' || a.agginitval
                    else ''
                  end
                  || case
                    when a.aggminitval is not null then ',' || E'\n'
                      || '  minitcond = ' || a.aggminitval
                    else ''
                  end
                  || case
                    when a.aggsortop != 0 then
                      ',' || E'\n' || '  sortop = ' || o.oprname
                    else ''
                  end
                  || E'\n' || ')'
                else pg_get_functiondef(p.oid)
            end as definition
            , x.object_args as objargs
            from pg_proc p
            inner join pg_namespace n on (p.pronamespace = n.oid)
            left join pg_aggregate a on (p.oid = a.aggfnoid)
            left join pg_type tt on (a.aggtranstype = tt.oid)
            left join pg_type tm on (a.aggmtranstype = tm.oid)
            left join pg_operator o on (a.aggsortop = o.oid)
            cross join lateral pg_identify_object_as_address
            ( 'pg_catalog.pg_proc'::regclass::oid
            , p.oid
            , 0
            ) x
            where p.oid = any(%s::oid[])
        """,
            (oids,),
        )
        procedures: list[Procedure] = []
        for row in await cur.fetchall():
            procedures.append(Procedure.model_validate(row))
        if len(procedures) != len(oids):
            logger.warning(
                f"{len(oids)} oids were provided but only {len(procedures)} procedures were loaded"
            )
    return procedures


async def _load_descriptions(
    con: psycopg.AsyncConnection, catalog_id: int, classid: int, objids: list[int]
) -> list[ObjectDescription]:
    """Loads ObjectDescriptions from the semantic catalog
    Given a classid and a list of objids, this will load the corresponding
    ObjectDescriptions.
    Args:
        con: Asynchronous database connection to the catalog database.
        catalog_id: ID of the semantic catalog to search in.
        classid: PostgreSQL object class ID
        objids: list of PostgreSQL object IDs
    Returns:
        A list of ObjectDescription objects.
    """
    async with con.cursor(row_factory=dict_row) as cur:
        sql = SQL("""\
            select x.*
            from ai.{table} x
            where x.classid = %(classid)s
            and x.objid = any(%(objids)s)
            order by x.classid, x.objid
        """).format(
            table=Identifier(f"semantic_catalog_obj_{catalog_id}"),
        )
        await cur.execute(
            sql,
            dict(
                classid=classid,
                objids=objids,
            ),
        )
        results: list[ObjectDescription] = []
        for row in await cur.fetchall():
            results.append(ObjectDescription(**row))
        return results


async def load_objects(
    catalog_con: psycopg.AsyncConnection,
    target_con: psycopg.AsyncConnection,
    catalog_id: int,
    obj_desc: list[ObjectDescription],
    sample_size: int = 0,
) -> list[Table | View | Procedure]:
    """Load database objects based on their descriptions.

    Takes a list of object descriptions and loads the corresponding database objects
    (tables, views, procedures) with their metadata. Matches the descriptions with
    the loaded objects and attaches them. If sample_size is greater than 0, it also
    retrieves sample data for tables and views.

    Args:
        catalog_con: Connection to the semantic catalog database.
        target_con: Connection to the target database where the objects are defined.
        catalog_id: ID of the semantic catalog to use for retrieving descriptions.
        obj_desc: List of object descriptions to match with database objects.
        sample_size: Number of sample rows to retrieve from tables and views (default: 0).
            If 0, no sample data is retrieved.

    Returns:
        A list of database objects (Tables, Views, Procedures) with metadata and descriptions.

    Raises:
        ValueError: If an unknown object type is encountered.
    """
    t: set[int] = set()  # distinct objid
    v: set[int] = set()  # distinct objid
    p: set[int] = set()  # distinct objid
    td: dict[int, ObjectDescription] = {}  # objid -> OD
    cd: dict[tuple[int, int], ObjectDescription] = {}  # (objid, objsubid) -> OD
    vd: dict[int, ObjectDescription] = {}  # objid -> OD
    pd: dict[int, ObjectDescription] = {}  # objid -> OD
    classid: int = -1
    tv: set[int] = set()  # all tables and views objid
    for od in obj_desc:
        match od.objtype:
            case "table":
                t.add(od.objid)
                td[od.objid] = od
                tv.add(od.objid)
                classid = od.classid
            case "view":
                v.add(od.objid)
                vd[od.objid] = od
                tv.add(od.objid)
                classid = od.classid
            case "procedure":
                p.add(od.objid)
                pd[od.objid] = od
            case "table column":
                t.add(od.objid)
                cd[(od.objid, od.objsubid)] = od
                tv.add(od.objid)
                classid = od.classid
            case "view column":
                v.add(od.objid)
                cd[(od.objid, od.objsubid)] = od
                tv.add(od.objid)
                classid = od.classid
            case _:
                raise ValueError(f"unknown object type {od.objtype}")
    # make sure all the descriptions for each table/view are loaded
    for od in await _load_descriptions(
        catalog_con, catalog_id, classid, [o for o in tv]
    ):
        match od.objtype:
            case "table":
                td[od.objid] = od
            case "view":
                vd[od.objid] = od
            case "table column":
                cd[(od.objid, od.objsubid)] = od
            case "view column":
                vd[od.objid] = od
            case _:
                raise ValueError(f"unknown object type {od.objtype}")
    tables = await load_tables(target_con, list(t), sample_size) if len(t) > 0 else []
    views = await load_views(target_con, list(v), sample_size) if len(v) > 0 else []
    procedures = await load_procedures(target_con, list(p)) if len(p) > 0 else []
    for table in tables:
        d = td.get(table.objid, None)
        table.description = d
        table.id = d.id if d is not None else -1
        if table.columns:
            for column in table.columns:
                column.description = cd.get((column.objid, column.objsubid), None)
    for view in views:
        d = vd.get(view.objid, None)
        view.description = d
        view.id = d.id if d is not None else -1
        if view.columns:
            for column in view.columns:
                column.description = cd.get((column.objid, column.objsubid), None)
    for procedure in procedures:
        d = pd.get(procedure.objid, None)
        procedure.description = d
        procedure.id = d.id if d is not None else -1
    results: list[Table | View | Procedure] = []
    results.extend(tables)
    results.extend(views)
    results.extend(procedures)
    return results



import os
import subprocess
from pathlib import Path, PosixPath

import psycopg
import pytest

import pgai

# skip tests in this module if disabled
enable_dump_restore_tests = os.getenv("ENABLE_DUMP_RESTORE_TESTS")
if enable_dump_restore_tests == "0":
    pytest.skip(allow_module_level=True)


USER = "jane"  # NOT a superuser


def db_url(user: str, dbname: str) -> str:
    return f"postgres://{user}@127.0.0.1:5432/{dbname}"


def where_am_i() -> str:
    if "WHERE_AM_I" in os.environ and os.environ["WHERE_AM_I"] == "docker":
        return "docker"
    return "host"


def docker_dir() -> str:
    return str(
        PosixPath("/").joinpath(
            "pgai", "projects", "pgai", "db", "tests", "dump_restore"
        )
    )


def host_dir() -> Path:
    return Path(__file__).parent.absolute()


def create_user(user: str) -> None:
    with psycopg.connect(
        db_url(user="postgres", dbname="postgres"), autocommit=True
    ) as con:
        with con.cursor() as cur:
            cur.execute(
                """
                select count(*) > 0
                from pg_catalog.pg_roles
                where rolname = %s
            """,
                (user,),
            )
            exists: bool = cur.fetchone()[0]
            if not exists:
                cur.execute(f"create user {user}")  # NOT a superuser


def create_database(dbname: str) -> None:
    with psycopg.connect(
        db_url(user="postgres", dbname="postgres"), autocommit=True
    ) as con:
        with con.cursor() as cur:
            cur.execute(f"drop database if exists {dbname} with (force)")
            cur.execute(f"create database {dbname} with owner {USER}")


def dump_db() -> None:
    host_dir().joinpath("dump.sql").unlink(missing_ok=True)
    cmd = " ".join(
        [
            "pg_dump -Fp --no-comments",
            f'''-d "{db_url(USER, "src")}"''',
            f"""-f {docker_dir()}/dump.sql""",
        ]
    )
    if where_am_i() != "docker":
        cmd = f"docker exec -w {docker_dir()} pgai-db {cmd}"
    subprocess.run(cmd, check=True, shell=True, env=os.environ, cwd=str(host_dir()))


def restore_db() -> None:
    cmd = " ".join(
        [
            "psql",
            f'''-d "{db_url(USER, "dst")}"''',
            "-v VERBOSITY=verbose",
            f"-f {docker_dir()}/dump.sql",
        ]
    )
    if where_am_i() != "docker":
        cmd = f"docker exec -w {docker_dir()} pgai-db {cmd}"
    subprocess.run(cmd, check=True, shell=True, env=os.environ, cwd=str(host_dir()))


def snapshot_db(dbname: str) -> None:
    host_dir().joinpath(f"{dbname}.snapshot").unlink(missing_ok=True)
    cmd = " ".join(
        [
            "psql",
            f'''-d "{db_url("postgres", dbname)}"''',
            "-v ON_ERROR_STOP=1",
            "-X",
            f"-o {docker_dir()}/{dbname}.snapshot",
            f"-f {docker_dir()}/snapshot.sql",
        ]
    )
    if where_am_i() != "docker":
        cmd = f"docker exec -w {docker_dir()} pgai-db {cmd}"
    subprocess.run(cmd, check=True, shell=True, env=os.environ, cwd=str(host_dir()))
    subprocess.run(cmd, check=True, shell=True, env=os.environ, cwd=str(host_dir()))


def init_src() -> None:
    pgai.install(db_url(user=USER, dbname="src"))
    cmd = " ".join(
        [
            "psql",
            f'''-d "{db_url(USER, "src")}"''',
            "-v ON_ERROR_STOP=1",
            f"-f {docker_dir()}/init.sql",
        ]
    )
    if where_am_i() != "docker":
        cmd = f"docker exec -w {docker_dir()} pgai-db {cmd}"
    subprocess.run(cmd, check=True, shell=True, env=os.environ, cwd=str(host_dir()))


def read_file(filename: str) -> str:
    with open(filename) as f:
        return f.read()


def after_dst() -> None:
    cmd = " ".join(
        [
            "psql",
            f'''-d "{db_url(USER, "dst")}"''',
            "-v ON_ERROR_STOP=1",
            f"-f {docker_dir()}/after.sql",
        ]
    )
    if where_am_i() != "docker":
        cmd = f"docker exec -w {docker_dir()} pgai-db {cmd}"
    subprocess.run(cmd, check=True, shell=True, env=os.environ, cwd=str(host_dir()))


def count_vectorizers() -> int:
    with psycopg.connect(db_url(user=USER, dbname="dst"), autocommit=True) as con:
        with con.cursor() as cur:
            cur.execute("select count(*) from ai.vectorizer")
            count: int = cur.fetchone()[0]
            return count


def test_dump_restore():
    create_user(USER)
    create_user("ethel")
    create_database("src")
    create_database("dst")
    init_src()
    snapshot_db("src")
    dump_db()
    restore_db()
    snapshot_db("dst")
    src = read_file(str(host_dir().joinpath("src.snapshot")))
    dst = read_file(str(host_dir().joinpath("dst.snapshot")))
    assert dst == src
    after_dst()  # make sure we can USE the restored db
    assert count_vectorizers() == 2



"""Vectorizer package for creating and managing vector embeddings in the semantic catalog.

This package provides functionality for converting textual content in the semantic
catalog into vector embeddings using various embedding models. It supports multiple
embedding providers including SentenceTransformers, Ollama, and OpenAI.

The vectorizer package is used to generate embeddings for database objects,
SQL examples, and facts in the semantic catalog, enabling vector similarity search.
"""

from .vectorizer import (
    EmbeddingConfig,
    OllamaConfig,
    OpenAIConfig,
    SentenceTransformersConfig,
    embedding_config_from_dict,
    vectorize,
    vectorize_query,
)

__all__ = [
    "SentenceTransformersConfig",
    "OllamaConfig",
    "OpenAIConfig",
    "EmbeddingConfig",
    "embedding_config_from_dict",
    "vectorize",
    "vectorize_query",
]



import os

import psycopg
import pytest
from psycopg.rows import dict_row

# skip tests in this module if disabled
enable_vectorizer_tests = os.getenv("ENABLE_VECTORIZER_TESTS")
if enable_vectorizer_tests == "0":
    pytest.skip(allow_module_level=True)


def db_url(user: str) -> str:
    return f"postgres://{user}@127.0.0.1:5432/test"


def test_worker_tracking_connection():
    with psycopg.connect(db_url("test")) as con:
        with con.cursor(row_factory=dict_row) as cur:
            row = cur.execute(
                "select ai._worker_start('0.1.1', interval '2 second') as worker_id"
            ).fetchone()
            worker_id = row["worker_id"]

            cur.execute("select ai._worker_heartbeat(%s, 1, 0, null)", (worker_id,))
            cur.execute(
                "select id, version, heartbeat_count, error_count, last_error_at, last_error_message, success_count from ai.vectorizer_worker_process where id = %s",
                (worker_id,),
            )
            row = cur.fetchone()
            assert row is not None
            assert row == {
                "id": worker_id,
                "version": "0.1.1",
                "heartbeat_count": 1,
                "error_count": 0,
                "last_error_at": None,
                "last_error_message": None,
                "success_count": 1,
            }
        con.commit()

        with con.cursor(row_factory=dict_row) as cur:
            cur.execute(
                "select ai._worker_heartbeat(%s, 0, 3, 'error 1')", (worker_id,)
            )
            cur.execute(
                "select id, version, heartbeat_count, error_count, last_error_at, last_error_message, last_heartbeat, success_count from ai.vectorizer_worker_process where id = %s",
                (worker_id,),
            )
            row = cur.fetchone()
            assert row is not None
            assert row["last_error_at"] is not None
            last_error_at = row["last_error_at"]
            assert row == {
                "id": worker_id,
                "version": "0.1.1",
                "heartbeat_count": 2,
                "error_count": 3,
                "last_error_at": last_error_at,
                "last_error_message": "error 1",
                "last_heartbeat": last_error_at,
                "success_count": 1,
            }
        con.commit()

        with con.cursor(row_factory=dict_row) as cur:
            cur.execute("select ai._worker_heartbeat(%s,1, 0, null)", (worker_id,))
            cur.execute(
                "select id, version, heartbeat_count, error_count, last_error_at, last_error_message, last_heartbeat, success_count from ai.vectorizer_worker_process where id = %s",
                (worker_id,),
            )
            row = cur.fetchone()
            assert row is not None
            last_heartbeat = row["last_heartbeat"]
            assert last_heartbeat > last_error_at
            assert row == {
                "id": worker_id,
                "version": "0.1.1",
                "heartbeat_count": 3,
                "error_count": 3,
                "last_error_at": last_error_at,
                "last_error_message": "error 1",
                "last_heartbeat": last_heartbeat,
                "success_count": 2,
            }
        con.commit()

        with con.cursor(row_factory=dict_row) as cur:
            cur.execute(
                "select ai._worker_heartbeat(%s, 0, 1, 'error 2')", (worker_id,)
            )
            cur.execute(
                "select id, version, heartbeat_count, error_count, last_error_at, last_error_message, last_heartbeat, success_count from ai.vectorizer_worker_process where id = %s",
                (worker_id,),
            )
            row = cur.fetchone()
            assert row is not None
            assert row["last_error_at"] > last_error_at
            last_error_at = row["last_error_at"]
            assert row == {
                "id": worker_id,
                "version": "0.1.1",
                "heartbeat_count": 4,
                "error_count": 4,
                "last_error_at": last_error_at,
                "last_error_message": "error 2",
                "last_heartbeat": last_error_at,
                "success_count": 2,
            }
        con.commit()


def test_worker_tracking_progress():
    with psycopg.connect(db_url("test")) as con:
        with con.cursor(row_factory=dict_row) as cur:
            cur.execute("drop table if exists blog")
            cur.execute("""
                create table blog
                ( id int not null generated always as identity
                , title text not null
                , published timestamptz
                , body text not null
                , drop_me text
                , primary key (title, published)
                )
            """)

            cur.execute("""
            select ai.create_vectorizer
                ( 'blog'::regclass
                , loading => ai.loading_column('body')
                , embedding=>ai.embedding_openai('text-embedding-3-small', 3)
                , chunking=>ai.chunking_character_text_splitter(128, 10)
                , formatting=>ai.formatting_python_template('title: $title published: $published $chunk')
                ) as vectorizer_id;
            """)
            vectorizer_id = cur.fetchone()["vectorizer_id"]

            row = cur.execute(
                "select ai._worker_start('0.1.1', interval '2 second') as worker_id"
            ).fetchone()
            worker_id = row["worker_id"]

            cur.execute(
                "select ai._worker_progress(%s, %s, 1, null)",
                (worker_id, vectorizer_id),
            )
            cur.execute(
                "select vectorizer_id, last_success_at, last_error_at, last_error_message, last_success_process_id, last_error_process_id, success_count, error_count from ai.vectorizer_worker_progress where vectorizer_id = %s",
                (vectorizer_id,),
            )
            row = cur.fetchone()
            assert row is not None
            assert row["vectorizer_id"] == vectorizer_id
            assert row["last_success_at"] is not None
            assert row["last_error_at"] is None
            assert row["last_error_message"] is None
            assert row["last_success_process_id"] == worker_id
            assert row["last_error_process_id"] is None
            assert row["success_count"] == 1
            assert row["error_count"] == 0

            cur.execute(
                "select ai._worker_progress(%s, %s, 0, 'error 1')",
                (worker_id, vectorizer_id),
            )
            cur.execute(
                "select vectorizer_id, last_success_at, last_error_at, last_error_message, last_success_process_id, last_error_process_id, success_count, error_count from ai.vectorizer_worker_progress where vectorizer_id = %s",
                (vectorizer_id,),
            )
            row = cur.fetchone()
            assert row is not None
            assert row["vectorizer_id"] == vectorizer_id
            assert row["last_success_at"] is not None
            assert row["last_error_at"] is not None
            last_error_at = row["last_error_at"]
            assert row["last_error_message"] == "error 1"
            assert row["last_success_process_id"] == worker_id
            assert row["last_error_process_id"] == worker_id
            assert row["success_count"] == 1
            assert row["error_count"] == 1
        con.commit()

        with con.cursor(row_factory=dict_row) as cur:
            cur.execute(
                "select ai._worker_progress(%s, %s, 1, null)",
                (worker_id, vectorizer_id),
            )
            cur.execute(
                "select vectorizer_id, last_success_at, last_error_at, last_error_message, last_success_process_id, last_error_process_id, success_count, error_count from ai.vectorizer_worker_progress where vectorizer_id = %s",
                (vectorizer_id,),
            )
            row = cur.fetchone()
            assert row is not None
            assert row["vectorizer_id"] == vectorizer_id
            assert row["last_success_at"] > last_error_at
            assert row["last_error_at"] == last_error_at
            assert row["last_error_message"] == "error 1"
            assert row["last_success_process_id"] == worker_id
            assert row["last_error_process_id"] == worker_id
            assert row["success_count"] == 2
            assert row["error_count"] == 1
        con.commit()



import os

import psycopg
import pytest

# skip tests in this module if disabled
enable_vectorizer_tests = os.getenv("ENABLE_VECTORIZER_TESTS")
if enable_vectorizer_tests == "0":
    pytest.skip(allow_module_level=True)


def db_url(user: str) -> str:
    return f"postgres://{user}@127.0.0.1:5432/test"


def test_grant_to():
    tests = [
        (
            "",
            """
            select ai.grant_to()
            """,
            [],
        ),
        (
            "SET ai.grant_to_default='bob, barry, sal'",  # test spaces
            """
            select ai.grant_to()
            """,
            ["bob", "barry", "sal"],
        ),
        (
            "",
            """
            select ai.grant_to('ted')
            """,
            ["ted"],
        ),
        (
            "SET ai.grant_to_default='bob,barry,sal'",
            """
            select ai.grant_to('may')
            """,
            ["barry", "may", "sal", "bob"],
        ),
        (
            "SET ai.grant_to_default='bob , barry , sal '",
            """
            select ai.grant_to(variadic array[]::name[])
            """,
            ["barry", "sal", "bob"],
        ),
        (
            "",
            """
            select ai.grant_to(variadic array[]::name[])
            """,
            [],
        ),
    ]
    with psycopg.connect(db_url("test")) as con:
        with con.cursor() as cur:
            for setup, query, expected in tests:
                cur.execute(setup)
                cur.execute(query)
                actual = cur.fetchone()[0]
                assert actual.sort() == expected.sort()



import asyncio
import os
from pathlib import Path

import psycopg
import pytest
from psycopg.sql import SQL, Identifier

import pgai
from pgai.semantic_catalog import semantic_catalog
from pgai.semantic_catalog.vectorizer import (
    OllamaConfig,
    OpenAIConfig,
    SentenceTransformersConfig,
    vectorize,
)
from tests.semantic_catalog.utils import PostgresContainer

DATABASE = "sc_03"


@pytest.fixture(scope="module")
def event_loop():
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="module", autouse=True)
async def setup_database(container: PostgresContainer) -> None:
    container.drop_database(DATABASE, force=True)
    container.create_database(DATABASE)
    pgai.install(container.connection_string(database=DATABASE))
    script = (
        Path(__file__)
        .parent.joinpath("data", "descriptions_to_vectorize.sql")
        .read_text()
    )
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        # create a default semantic catalog
        sc = await semantic_catalog.create(con)
        # add a second embedding config
        await sc.add_embedding(
            con,
            OpenAIConfig.create(
                model="text-embedding-3-small",
                dimensions=1536,
            ),
        )
        # add a third embedding config
        await sc.add_embedding(
            con,
            OllamaConfig.create(
                model="nomic-embed-text",
                dimensions=768,
            ),
        )
        # load descriptions into the semantic catalog
        async with con.cursor() as cur:
            await cur.execute(script)  # pyright: ignore [reportArgumentType]


async def vectorize_test(
    container: PostgresContainer, emb_name: str, emb_type: type
) -> None:
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        async with con.cursor() as cur:
            await cur.execute("""\
                    select count(*) > 0
                    from ai.semantic_catalog_obj_1
                """)
            row = await cur.fetchone()
            assert row is not None
            actual: bool = row[0]
            assert actual > 0, "found no obj rows"
            await cur.execute("""\
                    select count(*) > 0
                    from ai.semantic_catalog_sql_1
                """)
            row = await cur.fetchone()
            assert row is not None
            actual: bool = row[0]
            assert actual > 0, "found no sql rows"
            await cur.execute("""\
                    select count(*) > 0
                    from ai.semantic_catalog_fact_1
                """)
            row = await cur.fetchone()
            assert row is not None
            actual: bool = row[0]
            assert actual > 0, "found no fact rows"
        sc = await semantic_catalog.from_name(con, "default")
        config = await sc.get_embedding(con, emb_name)
        assert config is not None, f"embedding config {emb_name} not found"
        assert isinstance(
            config, emb_type
        ), f"embedding config {emb_name} not of type {emb_type}"
        async with con.cursor() as cur:
            await cur.execute(
                SQL("""\
                select count(*) > 0
                from ai.semantic_catalog_obj_1
                where {} is null
            """).format(Identifier(emb_name))
            )
            assert row is not None
            actual: bool = row[0]
            assert actual is True, f"found no obj rows with null vectors for {emb_name}"
            await cur.execute(
                SQL("""\
                select count(*) > 0
                from ai.semantic_catalog_sql_1
                where {} is null
            """).format(Identifier(emb_name))
            )
            assert row is not None
            actual: bool = row[0]
            assert actual is True, f"found no sql rows with null vectors for {emb_name}"
            await cur.execute(
                SQL("""\
                select count(*) > 0
                from ai.semantic_catalog_fact_1
                where {} is null
            """).format(Identifier(emb_name))
            )
            assert row is not None
            actual: bool = row[0]
            assert (
                actual is True
            ), f"found no fact rows with null vectors for {emb_name}"
        await vectorize(con, sc.id, emb_name, config)
        async with con.cursor() as cur:
            await cur.execute(
                SQL("""\
                select count(*) = 0
                from ai.semantic_catalog_obj_1
                where {} is null
            """).format(Identifier(emb_name))
            )
            assert row is not None
            actual: bool = row[0]
            assert actual is True, f"found obj rows with null vectors for {emb_name}"
            await cur.execute(
                SQL("""\
                select count(*) = 0
                from ai.semantic_catalog_sql_1
                where {} is null
            """).format(Identifier(emb_name))
            )
            assert row is not None
            actual: bool = row[0]
            assert actual is True, f"found sql rows with null vectors for {emb_name}"
            await cur.execute(
                SQL("""\
                select count(*) = 0
                from ai.semantic_catalog_fact_1
                where {} is null
            """).format(Identifier(emb_name))
            )
            assert row is not None
            actual: bool = row[0]
            assert actual is True, f"found fact rows with null vectors for {emb_name}"


async def test_vectorize_sentence_transformers(container: PostgresContainer) -> None:
    await vectorize_test(container, "emb1", SentenceTransformersConfig)


async def test_vectorize_openai(container: PostgresContainer) -> None:
    if "OPENAI_API_KEY" not in os.environ:
        pytest.skip("OPENAI_API_KEY is not set")
    await vectorize_test(container, "emb2", OpenAIConfig)


async def test_vectorize_ollama(container: PostgresContainer) -> None:
    if "OLLAMA_HOST" not in os.environ:
        pytest.skip("OLLAMA_HOST is not set")
    await vectorize_test(container, "emb3", OllamaConfig)



# ruff: noqa: E402

import sys
from importlib.util import find_spec

tiktoken = find_spec("tiktoken")
if tiktoken is None:
    print(
        "vectorizer-worker extra is not installed, please install it with `pip install pgai[vectorizer-worker]`",  # noqa: E501
        file=sys.stderr,
    )
    sys.exit(1)

from .create_vectorizer import CreateVectorizer
from .vectorizer import Executor, Vectorizer
from .worker import Worker

__all__ = [
    "Vectorizer",
    "Executor",
    "CreateVectorizer",
    "Worker",
]



from importlib.resources import files
from textwrap import dedent
from typing import Any

import psycopg
import semver
import structlog
from psycopg import sql as sql_lib

from .. import __version__

GUC_VECTORIZER_URL = "ai.external_functions_executor_url"

log = structlog.get_logger()


def _get_sql(vector_extension_schema: str) -> str:
    with files("pgai.data").joinpath("ai.sql").open(mode="r") as f:
        sql = f.read()
    sql = sql.replace("@extschema:vector@", vector_extension_schema)
    sql = sql.replace("__version__", __version__)
    return sql


def warn_if_pre_release() -> None:
    if semver.VersionInfo.parse(__version__).prerelease is not None:
        log.warning(
            dedent("""
            Installing pre-release version of pgai.

            This is unstable software and no upgrade path is guaranteed.

            Instead, install using the latest release in pip:
            https://pypi.org/project/pgai/
        """)
        )


def _get_guc_vectorizer_url_sql() -> sql_lib.SQL:
    return sql_lib.SQL("select pg_catalog.current_setting(%s, true) as val")


def _get_vector_extension_schema_sql() -> sql_lib.SQL:
    return sql_lib.SQL("""
        select n.nspname
        from pg_extension e
        join pg_namespace n on n.oid = e.extnamespace
        where e.extname = 'vector'
    """)


def get_ai_extension_version_sql() -> sql_lib.SQL:
    return sql_lib.SQL("select extversion from pg_extension where extname = 'ai'")


def raise_on_old_extension_version(result: Any | None) -> None:
    if result is None or result[0] is None:
        return  # no extension installed is fine

    if semver.VersionInfo.parse(result[0]) < semver.VersionInfo.parse("0.10.0"):
        raise Exception(
            f"The ai extension is outdated and must be upgraded. Installed version: {result[0]}, required version: 0.10.0"
        )


def _get_server_version_sql() -> sql_lib.SQL:
    return sql_lib.SQL(
        "select current_setting('server_version_num', true)::int / 10000"
    )


def verify_error_library_already_installed(
    error_from_result: psycopg.errors.DuplicateObject,
) -> bool:
    if error_from_result.diag.message_primary is None:
        return False
    return (
        "the pgai library has already been installed/upgraded"
        in error_from_result.diag.message_primary
    )


async def ainstall(
    db_url: str, vector_extension_schema: str | None = None, strict: bool = False
) -> None:
    """Asynchronously install the pgai library into a PostgreSQL database.

    Args:
        db_url: Database connection URL
        vector_extension_schema: Schema where the vector extension is installed if it
            doesn't exist. If None, then the vector extension will be installed in the
            default schema (default: None)
        strict: If False, ignore if library is already installed. If True,
            raise error (default: False)

    Raises:
        psycopg.errors.DuplicateObject: If library is already installed and
            strict=True
    """
    warn_if_pre_release()
    async with (
        await psycopg.AsyncConnection.connect(db_url, autocommit=True) as conn,
        conn.cursor() as cur,
        conn.transaction(),
    ):
        await cur.execute(_get_server_version_sql())
        result = await cur.fetchone()
        pg_version = int(result[0]) if result is not None else None
        if pg_version and pg_version < 15:
            raise RuntimeError(
                f"postgres {pg_version} is unsupported, pgai requires postgres version 15 or greater"
            )

        if vector_extension_schema is None:
            await conn.execute("CREATE EXTENSION IF NOT EXISTS vector")
        else:
            await conn.execute(
                sql_lib.SQL(
                    "CREATE EXTENSION IF NOT EXISTS vector WITH SCHEMA {}"
                ).format(sql_lib.Literal(vector_extension_schema))
            )

        await cur.execute(_get_vector_extension_schema_sql())
        result = await cur.fetchone()
        if result is None or result[0] is None:
            raise Exception("vector extension not installed")

        sql = _get_sql(result[0])

        # if we need to send a ping to an external url then
        # we need to install the ai extension
        await cur.execute(_get_guc_vectorizer_url_sql(), (GUC_VECTORIZER_URL,))
        result = await cur.fetchone()
        if result is not None and result[0] is not None:
            await conn.execute("CREATE EXTENSION IF NOT EXISTS ai cascade")

        # check if the ai extension is installed is at an acceptable version
        await cur.execute(get_ai_extension_version_sql())
        raise_on_old_extension_version(await cur.fetchone())

        try:
            await conn.execute(sql)  # type: ignore
        except psycopg.errors.DuplicateObject as error_from_result:
            # note the duplicate object error is raised in head.sql by a raise
            # that uses the 42710 error code.
            if not strict and verify_error_library_already_installed(error_from_result):
                pass
            else:
                raise error_from_result


def install(
    db_url: str, vector_extension_schema: str | None = None, strict: bool = False
) -> None:
    """Install the pgai library into a PostgreSQL database.

    Args:
        db_url: Database connection URL
        vector_extension_schema: Schema where the vector extension is installed if it
            doesn't exist. If None, then the vector extension will be installed in the
            default schema (default: None)
        strict: If False, ignore if library is already installed. If True,
            raise error (default: False)

    Raises:
        psycopg.errors.DuplicateObject: If library is already installed and
            strict=True
    """
    warn_if_pre_release()
    with (
        psycopg.connect(db_url, autocommit=True) as conn,
        conn.cursor() as cur,
    ):
        cur.execute(_get_server_version_sql())
        result = cur.fetchone()
        pg_version = int(result[0]) if result is not None else None
        if pg_version and pg_version < 15:
            raise RuntimeError(
                f"postgres {pg_version} is unsupported, pgai requires postgres version 15 or greater"
            )

        if vector_extension_schema is None:
            conn.execute("CREATE EXTENSION IF NOT EXISTS vector")
        else:
            conn.execute(
                sql_lib.SQL(
                    "CREATE EXTENSION IF NOT EXISTS vector WITH SCHEMA {}"
                ).format(sql_lib.Literal(vector_extension_schema))
            )

        cur.execute(_get_vector_extension_schema_sql())
        result = cur.fetchone()
        if result is None or result[0] is None:
            raise Exception("vector extension not installed")

        sql = _get_sql(result[0])

        # if we need to send a ping to an external url then
        # we need to install the ai extension
        cur.execute(_get_guc_vectorizer_url_sql(), (GUC_VECTORIZER_URL,))
        result = cur.fetchone()
        if result is not None and result[0] is not None:
            conn.execute("CREATE EXTENSION IF NOT EXISTS ai cascade")

        # check if the ai extension is installed is at an acceptable version
        cur.execute(get_ai_extension_version_sql())
        raise_on_old_extension_version(cur.fetchone())

        try:
            conn.execute(sql)  # type: ignore
        except psycopg.errors.DuplicateObject as error_from_result:
            # note the duplicate object error is raised in head.sql by a raise
            # that uses the 42710 error code.
            if not strict and verify_error_library_already_installed(error_from_result):
                pass
            else:
                raise error_from_result



import os

from ddtrace.trace import tracer


def configure_tracing():
    # Disable tracer by default
    if os.getenv("DD_TRACE_ENABLED", "false").lower() == "true":
        tracer.enabled = True
    else:
        tracer.enabled = False
